#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Fill autosummary entries """

import importlib
import inspect


def gen_autosum(source, target, module, autotype, dtype="all", ignore=None):
    '''
    Automatically write a sphinx-parsable file, adding a list of functions or
    classes to the autosummary method of sphinx, in place of the @autosum@
    keyword.

    Parameters
    ----------
    source : str
        Name of the input file, usually of the form "source.rst.in".
    target : str
        Name of the output file, usually "source.rst".
    module : str
        Name of the module on which autosummary should be performed.
    autotype : str
        Type of summary (normal for all, 'autofunction' or 'autoclass').
        Use 'autoall' for both functions and classes.
        Use 'full' to get both summary and detailed list.
    dtype : str, optional (default: all)
        Type of object that should be kept ('func' or 'class'), depending on
        `autotype`.
    ignore : list, optional (default: None)
        Names of the objects that should not be included in the summary.
    '''
    # load module and get content
    mod     = importlib.import_module(module)
    mod_dir = dir(mod)

    # set ignored classes
    ignore = [] if ignore is None else ignore

    # list classes and functions
    str_autosum = ''
    str_autoref = ('\nDetails\n'
                   '=======\n\n'
                   '.. module:: ' + module + '\n\n')

    for member in mod_dir:
        if not member.startswith('_') and not member in ignore:
            m = getattr(mod, member)
            keep = 1

            isfunc  = inspect.isfunction(m)
            isclass = inspect.isclass(m)

            currenttype = None

            if autotype in ("autofunction", "autoclass"):
                currenttype = autotype
            elif isfunc:
                currenttype = 'autofunction'
            elif isclass:
                currenttype = 'autoclass'

            if dtype == "func":
                keep *= isfunc
            elif dtype == "class":
                keep *= isclass
            else:
                keep *= isfunc + isclass

            if keep:
                if autotype in ("summary", "full"):
                    str_autosum += '    ' + module + '.' + member + '\n'
                else:
                    str_autosum += '\n.. {}:: {}\n'.format(currenttype, member)

                    if currenttype == 'autoclass':
                        str_autosum += '    :members:\n'

                if autotype == "full":
                    str_autoref += '\n.. {}:: {}\n'.format(currenttype, member)

                    if currenttype == 'autoclass':
                        str_autosum += '    :members:\n'

    # write to file
    done = False

    str_final = ''

    with open(source, "r") as rst_input:
            for line in rst_input:
                if line.find("@autosum@") != -1 and not done:
                    str_final += str_autosum + "\n"

                    if autotype == "full":
                        str_final += str_autoref + '\n'

                    done = True
                else:
                    str_final += line

    with open(target, "w") as main_rst:
        main_rst.write(str_final)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# test_mpi.py
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the main methods of the :mod:`~nngt.generation` module.
"""

import os
import unittest

import numpy as np

import nngt
from nngt.analysis import *
from nngt.lib.connect_tools import _compute_connections

from base_test import TestBasis, XmlHandler, network_dir
from test_generation import _distance_rule_theo, _distance_rule_exp
from tools_testing import foreach_graph


if os.environ.get("MPI"):
    nngt.set_config("mpi", True)


# -------- #
# Test MPI #
# -------- #

class TestMPI(TestBasis):

    '''
    Class testing the main methods of the :mod:`~nngt.generation` module.
    '''

    theo_prop = {
        "distance_rule": _distance_rule_theo,
    }

    exp_prop = {
        "distance_rule": _distance_rule_exp,
    }

    tolerance = 0.08

    @property
    def test_name(self):
        return "test_mpi"

    @unittest.skipIf(not nngt.get_config('mpi'), "Not using MPI.")
    def gen_graph(self, graph_name):
        di_instructions = self.parser.get_graph_options(graph_name)
        graph = nngt.generate(di_instructions)
        if nngt.on_master_process():
            graph.set_name(graph_name)
        return graph, di_instructions

    @foreach_graph
    def test_model_properties(self, graph, instructions, **kwargs):
        '''
        When generating graphs from on of the preconfigured models, check that
        the expected properties are indeed obtained.
        '''
        if nngt.get_config("backend") != "nngt" and nngt.on_master_process():
            graph_type = instructions["graph_type"]
            ref_result = self.theo_prop[graph_type](instructions)
            computed_result = self.exp_prop[graph_type](graph, instructions)

            if graph_type == 'distance_rule':
                # average degree
                self.assertTrue(
                    ref_result[0] == computed_result[0],
                    "Avg. deg. for graph {} failed:\nref = {} vs exp {}\
                    ".format(graph.name, ref_result[0], computed_result[0]))

                # average error on distance distribution
                sqd = np.square(
                    np.subtract(ref_result[1:], computed_result[1:]))
                avg_sqd = sqd / np.square(computed_result[1:])
                err = np.sqrt(avg_sqd).mean()
                tolerance = (self.tolerance if instructions['rule'] == 'lin'
                             else 0.25)

                self.assertTrue(err <= tolerance,
                    "Distance distribution for graph {} failed:\nerr = {} > {}\
                    ".format(graph.name, err, tolerance))
        elif nngt.get_config("backend") == "nngt":
            from mpi4py import MPI
            comm       = MPI.COMM_WORLD
            num_proc   = comm.Get_size()
            graph_type = instructions["graph_type"]
            ref_result = self.theo_prop[graph_type](instructions)
            computed_result = self.exp_prop[graph_type](graph, instructions)
            if graph_type == 'distance_rule':
                # average degree
                self.assertTrue(
                    ref_result[0] == computed_result[0] * num_proc,
                    "Avg. deg. for graph {} failed:\nref = {} vs exp {}\
                    ".format(graph.name, ref_result[0], computed_result[0]))
                # average error on distance distribution
                sqd = np.square(
                    np.subtract(ref_result[1:], computed_result[1:]))
                avg_sqd = sqd / np.square(computed_result[1:])
                err = np.sqrt(avg_sqd).mean()
                tolerance = (self.tolerance if instructions['rule'] == 'lin'
                             else 0.25)
                self.assertTrue(err <= tolerance,
                    "Distance distribution for graph {} failed:\nerr = {} > {}\
                    ".format(graph.name, err, tolerance))



# ---------- #
# Test suite #
# ---------- #

if nngt.get_config('mpi'):
    suite = unittest.TestLoader().loadTestsFromTestCase(TestMPI)

if __name__ == "__main__":
    unittest.main()



#-*- coding:utf-8 -*-

# test_gallery.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.


""" Testing main functions """

import os
import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
import pytest

import nngt

folder = os.path.dirname(__file__)
folder = folder if folder else "."

root   = os.path.abspath(folder + "/..")
struct = root + "/examples/graph_structure"
props  = root + "/examples/graph_properties"


def mock_show():
    pass


@pytest.mark.mpi_skip
def test_structure(monkeypatch):
    '''
    Test gallery for structure visualization.
    '''
    monkeypatch.setattr(plt, "show", mock_show)

    for _, _, files in os.walk(struct):
        for fname in files:
            if fname.endswith(".py"):
                exec(open(fname).read())


@pytest.mark.mpi_skip
def test_properties(monkeypatch):
    '''
    Test gallery for graph properties.
    '''
    monkeypatch.setattr(plt, "show", mock_show)

    for _, _, files in os.walk(props):
        for fname in files:
            if fname.endswith(".py"):
                exec(open(fname).read())


if __name__ == "__main__":
    class mptch:
        def setattr(*args):
            pass

    mp = mptch()

    test_structure(mp)
    test_properties(mp)



#-*- coding:utf-8 -*-
#
# io/loading_helpers.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Loading helpers """

import re
import types

import numpy as np

from ..lib.converters import (_np_dtype, _to_int, _to_string, _to_list,
                              _string_from_object)


__all__ = [
    "_cleanup_line",
    "_gen_convert",
    "_get_edges_elist",
    "_get_edges_gml",
    "_get_edges_neighbour",
    "_get_node_attr",
    "_get_notif",
    "_process_file",
]


# ----------------------- #
# Initial file processing #
# ----------------------- #

def _process_file(f, fmt, separator):
    if fmt == "gml":
        # format gml input to expected one
        lines = []

        for l in f.readlines():
            clean_line = _cleanup_line(l, separator)

            if clean_line.endswith("[") and len(clean_line) > 1:
                lines.append(clean_line[:-1].strip())
                lines.append("[")
            elif clean_line.endswith("]") and len(clean_line) > 1:
                lines.append(clean_line[:-1].strip())
                lines.append("]")
            else:
                lines.append(clean_line)

        return lines

    # otherwise just cleanup the lines
    return [_cleanup_line(line, separator) for line in f.readlines()]


# ---------------- #
# Graph properties #
# ---------------- #

def _format_notif(notif_name, notif_val):
    attr = (
        "node_attributes", "edge_attributes", "node_attr_types",
        "edge_attr_types"
    )
    if notif_name in attr:
        lst = notif_val[1:-1].split(", ")
        if lst != ['']:  # check empty string
            return [val.strip("'\"") for val in lst]
        else:
            return []
    elif notif_name == "size":
        return int(notif_val)
    elif notif_name == "directed":
        return False if notif_val in ("False", "0") else True
    else:
        return notif_val


def _get_notif(lines, notifier, attributes, fmt=None, atypes=None):
    di_notif = {
        "node_attributes": [], "edge_attributes": [], "node_attr_types": [],
        "edge_attr_types": [],
    }

    # special case for GML
    if fmt == "gml":
        start = 0

        for i, l in enumerate(lines):
            if l == "graph":
                start = i
                break

        # nodes
        nodes = [i for i, l in enumerate(lines) if l == "node" and i > start]
        num_nodes = len(nodes)

        di_notif["size"]  = num_nodes
        di_notif["nodes"] = nodes

        # node attributes
        diff = np.diff(nodes) - 4  # number of lines other than node spec

        num_nattr = diff[0]

        if not np.all(diff == num_nattr):
            raise RuntimeError("All nodes should have the same attributes.")

        if num_nattr > len(di_notif["node_attributes"]):
            for i in range(nodes[0] + 3, nodes[0] + num_nattr + 3):
                name = lines[i].split(" ")[0]
                di_notif["node_attributes"].append(name)
                # default type is object
                if atypes is not None:
                    di_notif["node_attr_types"].append(
                        _string_from_object(atypes.get(name, object)))
                else:
                    di_notif["node_attr_types"].append("object")

        # graph attributes
        for line in lines[:nodes[0]]:
            first_space = line.find(" ")
            key, val = line[:first_space], line[first_space + 1:]

            di_notif[key] = _format_notif(key, val)

        # edges
        edges = [i for i, l in enumerate(lines) if l == "edge" and i > start]

        di_notif["edges"] = edges

        diff = np.diff(edges) - 5  # number of lines other than edge spec

        num_eattr = diff[0]

        if not np.all(diff == num_eattr):
            raise RuntimeError("All edges should have the same attributes.")

        if num_eattr > len(di_notif["edge_attributes"]):
            for i in range(edges[0] + 4, edges[0] + num_eattr + 4):
                name = lines[i].split(" ")[0]
                di_notif["edge_attributes"].append(name)
                # default type is object
                if atypes is not None:
                    di_notif["edge_attr_types"].append(
                        _string_from_object(atypes.get(name, object)))
                else:
                    di_notif["edge_attr_types"].append("object")
    else:
        for line in lines:
            if line.startswith(notifier):
                idx_eq = line.find("=")
                notif_name = line[len(notifier):idx_eq]
                notif_val = line[idx_eq+1:]
                di_notif[notif_name] = _format_notif(notif_name, notif_val)
            else:
                break

        if attributes is not None:
            di_notif["edge_attributes"] = attributes

            if atypes is not None:
                for attr in attributes:
                    di_notif["edge_attr_types"].append(
                        _string_from_object(atypes.get(attr, object)))
            else:
                di_notif["edge_attributes"] = ["object"]*len(attributes)

    return di_notif


# ----- #
# Edges #
# ----- #

def _get_edges_neighbour(lst_lines, attributes, ignore, notifier, separator,
                         secondary, di_attributes, di_convert, **kwargs):
    '''
    Add edges and attributes to `edges` and `di_attributes` for the "neighbour"
    format.
    '''
    edges = []

    lst_lines = lst_lines[::-1]

    while lst_lines:
        line = lst_lines.pop()

        if line and not (line.startswith(notifier) or line.startswith(ignore)):
            len_first_delim = line.find(separator)

            if len_first_delim > 0:
                source = int(line[:len_first_delim])
                len_first_delim += 1

                if len_first_delim:
                    neighbours = line[len_first_delim:].split(separator)

                    for stub in neighbours:
                        content = stub.split(secondary)
                        target = int(content[0])

                        edges.append((source, target))

                        attr_val = content[1:] if len(content) > 1 else []

                        for name, val in zip(attributes, attr_val):
                            di_attributes[name].append(di_convert[name](val))

    return edges


def _get_edges_elist(lst_lines, attributes, ignore, notifier, separator,
                     secondary, di_attributes, di_convert, **kwargs):
    '''
    Add edges and attributes to `edges` and `di_attributes` for the "neighbour"
    format.
    '''
    edges = []

    lst_lines = lst_lines[::-1]

    while lst_lines:
        line = lst_lines.pop()

        if line and not (line.startswith(notifier) or line.startswith(ignore)):
            data = line.split(separator)
            source, target = int(data[0]), int(data[1])
            edges.append((source, target))

            # different ways of loading the attributes
            if len(data) == 3 and secondary in data[2]:  # secondary notifier
                attr_data = data[2].split(secondary)
                for name, val in zip(attributes, attr_data):
                    di_attributes[name].append(di_convert[name](val))
            elif len(data) == len(attributes) + 2:  # regular columns
                for name, val in zip(attributes, data[2:]):
                    di_attributes[name].append(di_convert[name](val))

    return edges


def _get_edges_gml(lst_lines, attributes, *args, di_attributes=None,
                   di_convert=None, di_notif=None):
    '''
    Add edges and attributes to `edges` and `di_attributes` for the "neighbour"
    format.
    '''
    edges = []

    edge_lines = di_notif["edges"]
    num_eattr  = len(di_attributes)

    for line_num in edge_lines:
        source = int(lst_lines[line_num + 2][7:])
        target = int(lst_lines[line_num + 3][7:])

        edges.append((source, target))

        for i, name in zip(range(num_eattr), attributes):
            lnum  = line_num + 4 + i
            start = lst_lines[lnum].find(" ") + 1
            attr  = lst_lines[lnum][start:]
            di_attributes[name].append(di_convert[name](attr))

    return edges


def _get_node_attr(di_notif, separator, fmt=None, lines=None, atypes=None):
    '''
    Return node attributes.

    For custom formats, attributes are stored under @na_{attr_name} in the
    file, so they are stored under the coresponding key in `di_notif`.

    For GML, need to get them from the nodes.
    '''
    di_nattr   = {}

    if fmt == "gml":
        node_lines = di_notif["nodes"]
        num_nattr  = node_lines[1] - node_lines[0] - 3  # lines other than attr

        has_types = len(di_notif["node_attr_types"]) == num_nattr

        if num_nattr:
            for line_num in node_lines:
                for i in range(num_nattr):
                    l = lines[line_num + i + 2]
                    first_space = l.find(" ")

                    # get attribute name and value
                    name = l[:first_space]
                    val  = l[first_space + 1:]

                    if name not in di_nattr:
                        di_nattr[name] = []

                    dtype = str if atypes is None else atypes.get(name, str)

                    if has_types:
                        dtype = _type_converter(di_notif["node_attr_types"][i])

                    di_nattr[name].append(dtype(val))
    else:
        nattr_name = {str("na_" + k): k for k in di_notif["node_attributes"]}
        nattr_type = di_notif["node_attr_types"]

        for k, s in di_notif.items():
            if k in nattr_name:
                attr  = nattr_name[k]
                idx   = di_notif["node_attributes"].index(attr)
                dtype = _np_dtype(nattr_type[idx])

                if dtype == object:
                    char = "" if s.strip()[0] != '"' else '"'
                    nsep = char + separator + char

                    if s.startswith(char):
                        s = s[1:]

                    if s.endswith(char):
                        s = s[:-1]

                    di_nattr[attr] = np.array(
                        [w for w in s.split(nsep)], dtype=dtype)
                else:
                    di_nattr[attr] = np.fromstring(s, sep=separator,
                                                   dtype=dtype)

    return di_nattr


# ---------- #
# Converters #
# ---------- #

def _gen_convert(attributes, attr_types, attributes_types=None):
    '''
    Generate a conversion dictionary that associates the right type to each
    attribute
    '''
    di_convert = {}

    if attributes and not attr_types:
        attr_types.extend(("string" for _ in attributes))

    for attr, attr_type in zip(attributes, attr_types):
        if attributes_types is not None and attr in attributes_types:
            # user defined converter
            di_convert[attr] = attributes_types[attr]
        elif attr_type in ("double", "float", "real"):
            di_convert[attr] = float
        elif attr_type in ("str", "string"):
            di_convert[attr] = lambda x: str(x).strip("\"'")
        elif attr_type in ("int", "integer"):
            di_convert[attr] = _to_int
        elif attr_type in ("lst", "list", "tuple", "array"):
            di_convert[attr] = _to_list
        elif attr_type == "object":
            di_convert[attr] = lambda x: x
        else:
            raise TypeError("Invalid attribute type: '{}'.".format(attr_type))

    return di_convert


def _cleanup_line(string, char):
    ''' Replace multiple occurrences of a separator and remove line ends '''
    pattern = char + '+'
    string = re.sub(pattern, char, string)
    return string.strip()



#-*- coding:utf-8 -*-
#
# analysis/graph_analysis.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Tools for graph analysis using the graph libraries """

import logging

import numpy as np
import scipy.sparse.linalg as spl

import nngt
import nngt.generation as ng

from nngt.lib import InvalidArgument, nonstring_container, is_integer
from nngt.lib.logger import _log_message
from . import clustering
from .activity_analysis import get_b2, get_firing_rate
from .bayesian_blocks import bayesian_blocks
from .clustering import *


logger = logging.getLogger(__name__)


# implemented function; import from proper backend is done at the bottom

__all__ = [
    "adjacency_matrix",
    "all_shortest_paths",
    "assortativity",
    "average_path_length",
    "betweenness",
    "betweenness_distrib",
    "binning",
	"closeness",
	"connected_components",
    "degree_distrib",
	"diameter",
    "node_attributes",
	"num_iedges",
	"reciprocity",
    "shortest_distance",
    "shortest_path",
    "small_world_propensity",
	"spectral_radius",
    "subgraph_centrality",
    "transitivity",
]


__all__.extend(clustering.__all__)


_backend_required = "Please install either networkx, igraph, or graph-tool " \
                    "to use this function."


# ---------------- #
# Graph properties #
# ---------------- #

def assortativity(g, degree, weights=None):
    '''
    Returns the assortativity of the graph.
    This tells whether nodes are preferentially connected together depending
    on their degree.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    degree : str
        The type of degree that should be considered.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.

    References
    ----------
    .. [newman-mixing-2003] M. E. J. Newman, "Mixing patterns in networks",
        Phys. Rev. E 67, 026126 (2003), see graph-tool below for links.
    .. [gt-assortativity] :gtdoc:`correlations.scalar_assortativity`
    .. [ig-assortativity] :igdoc:`assortativity`
    .. [nx-assortativity]
       :nxdoc:`algorithms.assortativity.degree_assortativity_coefficient`
    '''
    raise NotImplementedError(_backend_required)


def reciprocity(g):
    '''
    Calculate the edge reciprocity of the graph.

    The reciprocity is defined as the number of edges that have a reciprocal
    edge (an edge between the same nodes but in the opposite direction)
    divided by the total number of edges.
    This is also the probability for any given edge, that its reciprocal edge
    exists.
    By definition, the reciprocity of undirected graphs is 1.

    @todo: check whether we can get this for single nodes for all libraries.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.

    References
    ----------
    .. [wasserman-1994] S. Wasserman and K. Faust, "Social Network Analysis".
       (Cambridge University Press, Cambridge, 1994)
    .. [lopez-2007] Gorka Zamora-López, Vinko Zlatić, Changsong
       Zhou, Hrvoje Štefančić, and Jürgen Kurths "Reciprocity of networks with
       degree correlations and arbitrary degree sequences", Phys. Rev. E 77,
       016106 (2008) :doi:`10.1103/PhysRevE.77.016106`, :arxiv:`0706.3372`
    .. [gt-reciprocity] :gtdoc:`topology.edge_reciprocity`
    .. [ig-reciprocity] :igdoc:`reciprocity`
    .. [nx-reciprocity] :nxdoc:`algorithms.reciprocity.overall_reciprocity`
    '''
    raise NotImplementedError(_backend_required)


def transitivity(g, directed=True, weights=None):
    '''
    Same as :func:`~nngt.analysis.global_clustering`.
    '''
    return global_clustering(g, directed=directed, weights=weights)


def num_iedges(graph):
    '''
    Returns the number of inhibitory connections.

    For :class:`~nngt.Network` objects, this corresponds to the number of edges
    stemming from inhibitory nodes (given by
    :meth:`nngt.NeuralPop.inhibitory`).
    Otherwise, counts the edges where the type attribute is -1.
    '''
    if graph.is_network():
        inhib_nodes = graph.population.inhibitory

        return np.sum(graph.get_degrees("out", node_list=inhib_nodes))

    if "type" in graph.edge_attributes:
        return np.sum(graph.get_edge_attributes(name="type") < 0)

    return 0.


def connected_components(g, ctype=None):
    '''
    Returns the connected component to which each node belongs.

    .. versionadded:: 2.0

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    ctype : str, optional (default 'scc')
        Type of component that will be searched: either strongly connected
        ('scc', by default) or weakly connected ('wcc').

    Returns
    -------
    cc, hist : :class:`numpy.ndarray`
        The component associated to each node (`cc`) and the number of nodes in
        each of the component (`hist`).

    References
    ----------
    .. [gt-cc] :gtdoc:`topology.label_components`
    .. [ig-cc] :igdoc:`clusters`
    .. [nx-ucc] :nxdoc:`algorithms.components.connected_components`
    .. [nx-scc] :nxdoc:`algorithms.components.strongly_connected_components`
    .. [nx-wcc] :nxdoc:`algorithms.components.weakly_connected_components`
    '''
    raise NotImplementedError(_backend_required)


def diameter(g, directed=True, weights=False, is_connected=False):
    '''
    Returns the diameter of the graph.

    .. versionchanged:: 2.0
        Added `directed` and `is_connected` arguments.

    It returns infinity if the graph is not connected (strongly connected for
    directed graphs) unless `is_connected` is True, in which case it returns
    the longest existing shortest distance.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    directed : bool, optional (default: True)
        Whether to compute the directed diameter if the graph is directed.
        If False, then the graph is treated as undirected. The option switches
        to False automatically if `g` is undirected.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    is_connected : bool, optional (default: False)
        If False, check whether the graph is connected or not and return
        infinite diameter if graph is unconnected. If True, the graph is
        assumed to be connected.

    Warning
    -------
    For graph-tool, the [pseudo-diameter]_ is returned, which may sometime
    lead to inexact results.

    See also
    --------
    :func:`nngt.analysis.shortest_distance`

    References
    ----------
    .. [pseudo-diameter] http://en.wikipedia.org/wiki/Distance_%28graph_theory%29
    .. [dijkstra] E. Dijkstra, "A note on two problems in connexion with
       graphs." Numerische Mathematik, 1:269-271, 1959.
    .. [gt-diameter] :gtdoc:`topology.pseudo_diameter`
    .. [ig-diameter] :igdoc:`diameter`
    .. [nx-diameter] :nxdoc:`algorithms.distance_measures.diameter`
    .. [nx-dijkstra] :nxdoc:`algorithms.shortest_paths.weighted.all_pairs_dijkstra`
    '''
    raise NotImplementedError(_backend_required)


def small_world_propensity(g, directed=None, use_global_clustering=False,
                           use_diameter=False, weights=None,
                           combine_weights="mean", clustering="continuous",
                           lattice=None, random=None, return_deviations=False):
    r'''
    Returns the small-world propensity of the graph as first defined in
    [Muldoon2016]_.

    .. versionadded: 2.0

    .. math::

        \phi = 1 - \sqrt{\frac{\Pi_{[0, 1]}(\Delta_C^2) + \Pi_{[0, 1]}(\Delta_L^2)}{2}}

    with :math:`\Delta_C` the clustering deviation, i.e. the relative global or
    average clustering of `g` compared to two reference graphs

    .. math::

        \Delta_C = \frac{C_{latt} - C_g}{C_{latt} - C_{rand}}

    and :math:`Delta_L` the deviation of the average path length or diameter,
    i.e. the relative average path length of `g` compared to that of the
    reference graphs

    .. math::

        \Delta_L = \frac{L_g - L_{rand}}{L_{latt} - L_{rand}}.

    In both cases, *latt* and *rand* refer to the equivalent lattice and
    Erdos-Renyi (ER) graphs obtained by rewiring `g` to obtain respectively the
    highest and lowest combination of clustering and average path length.

    Both deviations are clipped to the [0, 1] range in case some graphs have a
    higher clustering than the lattice or a lower average path length than the
    ER graph.

    Parameters
    ----------
    g : :class:`~nngt.Graph` object
        Graph to analyze.
    directed : bool, optional (default: True)
        Whether to compute the directed clustering if the graph is directed.
        If False, then the graph is treated as undirected. The option switches
        to False automatically if `g` is undirected.
    use_global_clustering : bool, optional (default: True)
        If False, then the average local clustering is used instead of the
        global clustering.
    use_diameter : bool, optional (default: False)
        Use the diameter instead of the average path length to have more global
        information. Ccan also be much faster in some cases, especially using
        graph-tool as the backend.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.
    clustering : str, optional (default: 'continuous')
        Method used to compute the weighted clustering coefficients, either
        'barrat' [Barrat2004]_, 'continuous' (recommended), or 'onnela'
        [Onnela2005]_.
    lattice : :class:`nngt.Graph`, optional (default: generated from `g`)
        Lattice to use as reference (since its generation is deterministic,
        enables to avoid multiple generations when running the algorithm
        several times with the same graph)
    random : :class:`nngt.Graph`, optional (default: generated from `g`)
        Random graph to use as reference. Can be useful for reproducibility or
        for very sparse graphs where ER algorithm would statistically lead to
        a disconnected graph.
    return_deviations : bool, optional (default: False)
        If True, the deviations are also returned, in addition to the
        small-world propensity.

    Note
    ----
    If `weights` are provided, the distance calculation uses the inverse of
    the weights.
    This implementation differs slightly from the `original implementation
    <https://github.com/KordingLab/nctpy>`_ as it can also use the global
    instead of the average clustering coefficient, the diameter instead of
    the avreage path length, and it is generalized to directed networks.

    References
    ----------
    .. [Muldoon2016] Muldoon, Bridgeford, Bassett. Small-World Propensity and
        Weighted Brain Networks. Sci Rep 2016, 6 (1), 22057.
        :doi:`10.1038/srep22057`, :arxiv:`1505.02194`.
    .. [Barrat2004] Barrat, Barthelemy, Pastor-Satorras, Vespignani. The
        Architecture of Complex Weighted Networks. PNAS 2004, 101 (11).
        :doi:`10.1073/pnas.0400087101`.
    .. [Onnela2005] Onnela, Saramäki, Kertész, Kaski. Intensity and Coherence
        of Motifs in Weighted Complex Networks. Phys. Rev. E 2005, 71 (6),
        065103. :doi:`10.1103/physreve.71.065103`, arxiv:`cond-mat/0408629`.

    Returns
    -------
    phi : float in [0, 1]
        The small-world propensity.
    delta_l : float
        The average path-length deviation (if `return_deviations` is True).
    delta_c : float
        The clustering deviation (if `return_deviations` is True).

    See also
    --------
    :func:`nngt.analysis.average_path_length`
    :func:`nngt.analysis.diameter`
    :func:`nngt.analysis.global_clustering`
    :func:`nngt.analysis.local_clustering`
    :func:`nngt.generation.lattice_rewire`
    :func:`nngt.generation.random_rewire`
    '''
    # special case for too sparse (unconnected) graphs
    if g.edge_nb() < g.node_nb():
        if return_deviations:
            return np.NaN, np.NaN, np.NaN

        return np.NaN

    # check graph directedness
    directed = g.is_directed() if directed is None else directed

    if g.is_directed() and not directed:
        g = g.to_undirected(combine_weights)

    # rewired graph
    latt = ng.lattice_rewire(g, weight=weights) if lattice is None else lattice
    rand = ng.random_rewire(g) if random is None else random

    # compute average path-length using the inverse of the weights
    inv_w, inv_wl, inv_wr = None, None, None

    if weights not in (None, False):
        inv_w = 1 / g.edge_attributes[weights]
        inv_wl = 1 / latt.edge_attributes[weights]
        inv_wr = 1 / rand.edge_attributes[weights]

    l_latt, l_rand, l_g = None, None, None

    if use_diameter:
        l_latt = diameter(latt, directed=directed, weights=inv_wl)
        l_rand = diameter(rand, directed=directed, weights=inv_wr)
        l_g    = diameter(g, directed=directed, weights=inv_w)
    else:
        l_latt = average_path_length(latt, directed=directed, weights=inv_wl)
        l_rand = average_path_length(rand, directed=directed, weights=inv_wr)
        l_g    = average_path_length(g, directed=directed, weights=inv_w)

    # compute clustering
    c_latt, c_rand, c_g = None, None, None

    if use_global_clustering:
        c_latt = global_clustering(
            latt, directed=directed, weights=weights, method=clustering)

        c_rand = global_clustering(
            rand, directed=directed, weights=weights, method=clustering)

        c_g = global_clustering(
            g, directed=directed, weights=weights, method=clustering)
    else:
        c_latt = np.average(local_clustering(
            latt, directed=directed, weights=weights, method=clustering))

        c_rand = np.average(local_clustering(
            rand, directed=directed, weights=weights, method=clustering))

        c_g = np.average(local_clustering(
            g, directed=directed, weights=weights, method=clustering))

    # compute deltas
    delta_l = (l_g - l_rand) / (l_latt - l_rand) if l_latt != l_rand \
              else float(l_g > l_rand)
    delta_c = (c_latt - c_g) / (c_latt - c_rand)

    if np.isinf(l_rand):
        _log_message(logger, "WARNING", 'Randomized graph was unconnected.')

    if return_deviations:
        return 1 - np.sqrt(
            0.5*(np.clip(delta_l, 0, 1)**2 + np.clip(delta_c, 0, 1)**2)), \
            delta_l, delta_c
    else:
        return 1 - np.sqrt(
            0.5*(np.clip(delta_l, 0, 1)**2 + np.clip(delta_c, 0, 1)**2))


def shortest_path(g, source, target, directed=True, weights=None):
    '''
    Returns a shortest path between `source`and `target`.
    The algorithms returns an empty list if there is no path between the nodes.

    .. versionadded:: 2.0

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    source : int
        Node from which the path starts.
    target : int
        Node where the path ends.
    directed : bool, optional (default: True)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str or array, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.

    Returns
    -------
    path : array of ints
        Order of the nodes making up the path from `source` to `target`.

    References
    ----------
    .. [gt-sd] :gtdoc:`topology.shortest_distance`
    .. [ig-sp] :igdoc:`shortest_paths`
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.generic.shortest_path`
    '''
    raise NotImplementedError(_backend_required)


def all_shortest_paths(g, source, target, directed=True, weights=None):
    '''
    Yields all shortest paths from `source` to `target`.
    The algorithms returns an empty generator if there is no path between the
    nodes.

    .. versionadded:: 2.0

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    source : int
        Node from which the paths starts.
    target : int, optional (default: all nodes)
        Node where the paths ends.
    directed : bool, optional (default: True)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str or array, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.

    Returns
    -------
    all_paths : generator
        Generator yielding paths as lists of ints.

    References
    ----------
    .. [gt-sd] :gtdoc:`topology.all_shortest_paths`
    .. [ig-sp] :igdoc:`get_all_shortest_paths`
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.generic.all_shortest_paths`
    '''
    raise NotImplementedError(_backend_required)


def shortest_distance(g, sources=None, targets=None, directed=True,
                      weights=None):
    '''
    Returns the length of the shortest paths between `sources`and `targets`.
    The algorithms return infinity if there are no paths between nodes.

    .. versionadded:: 2.0

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    sources : list of nodes, optional (default: all)
        Nodes from which the paths must be computed.
    targets : list of nodes, optional (default: all)
        Nodes to which the paths must be computed.
    directed : bool, optional (default: True)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str or array, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.

    Returns
    -------
    distance : float, or 1d/2d numpy array of floats
        Distance (if single source and single target) or distance array.
        For multiple sources and targets, the shape of the matrix is (S, T),
        with S the number of sources and T the number of targets; for a single
        source or target, return a 1d-array of length T or S.

    References
    ----------
    .. [gt-sd] :gtdoc:`topology.shortest_distance`
    .. [ig-sp] :igdoc:`shortest_paths`
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.weighted.multi_source_dijkstra`
    '''
    raise NotImplementedError(_backend_required)


def average_path_length(g, sources=None, targets=None, directed=None,
                        weights=None, unconnected=False):
    r'''
    Returns the average shortest path length between `sources` and `targets`.
    The algorithms raises an error if all nodes are not connected unless
    `unconnected` is set to True.

    .. versionadded:: 2.0

    The average path length is defined as

    .. math::

       L = \frac{1}{N_p} \sum_{u,v} d(u, v),

    where :math:`N_p` is the number of paths between `sources` and `targets`,
    and :math:`d(u, v)` is the shortest path distance from u to v.

    If `sources` and `targets` are both None, then the total number of paths is
    :math:`N_p = N(N - 1)`, with :math:`N` the number of nodes in the graph.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    sources : list of nodes, optional (default: all)
        Nodes from which the paths must be computed.
    targets : list of nodes, optional (default: all)
        Nodes to which the paths must be computed.
    directed : bool, optional (default: True)
        Whether the edges should be considered as directed or not
        (automatically set to False if `g` is undirected).
    weights : str, optional (default: binary)
        Whether to use weighted edges to compute the distances. By default,
        all edges are considered to have distance 1.
    unconnected : bool, optional (default: False)
        If set to true, ignores unconnected nodes and returns the average path
        length of the existing paths.

    References
    ----------
    .. [gt-sd] :gtdoc:`topology.shortest_distance`
    .. [ig-sp] :igdoc:`shortest_paths`
    .. [nx-sp] :nxdoc:`algorithms.shortest_paths.generic.average_shortest_path_length`

    See also
    --------
    :func:`nngt.analysis.shortest_distance`
    '''
    raise NotImplementedError(_backend_required)


# ------------ #
# Centralities #
# ------------ #

def closeness(g, weights=None, nodes=None, mode="out", harmonic=False,
              default=np.NaN):
    r'''
    Returns the closeness centrality of some `nodes`.

    .. versionadded:: 2.0

    Closeness centrality of a node `u` is defined, for the harmonic version,
    as the sum of the reciprocal of the shortest path distance :math:`d_{uv}`
    from `u` to the N - 1 other nodes in the graph (if `mode` is "out",
    reciprocally :math:`d_{vu}`, the distance to `u` from another node v,
    if `mode` is "in"):

    .. math::

        C(u) = \frac{1}{N - 1} \sum_{v \neq u} \frac{1}{d_{uv}},

    or, using the arithmetic definition, as the reciprocal of the
    average shortest path distance to/from `u` over to all other nodes:

    .. math::

        C(u) = \frac{n - 1}{\sum_{v \neq u} d_{uv}},

    where `d_{uv}` is the shortest-path distance from `u` to `v`,
    and `n` is the number of nodes in the component.

    By definition, the distance is infinite when nodes are not connected by
    a path in the harmonic case (such that :math:`\frac{1}{d(v, u)} = 0`),
    while the distance itself is taken as zero for unconnected nodes in the
    first equation.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    nodes : list, optional (default: all nodes)
        The list of nodes for which the clutering will be returned
    mode : str, optional (default: "out")
        For directed graphs, whether the distances are computed from ("out") or
        to ("in") each of the nodes.
    harmonic : bool, optional (default: False)
        Whether the arithmetic (default) or the harmonic (recommended) version
        of the closeness should be used.

    Returns
    -------
    c : :class:`numpy.ndarray`
        The list of closeness centralities, on per node.

    .. warning ::
        For compatibility reasons (harmonic closeness is not implemented for
        igraph), the arithmetic version is used by default; however, it is
        recommended to use the harmonic version instead whenever possible.

    References
    ----------
    .. [gt-closeness] :gtdoc:`centrality.closeness`
    .. [ig-closeness] :igdoc:`closeness`
    .. [nx-harmonic] :nxdoc:`algorithms.centrality.harmonic_centrality`
    .. [nx-closeness] :nxdoc:`algorithms.centrality.closeness_centrality`
    '''
    raise NotImplementedError(_backend_required)


def betweenness(g, btype="both", weights=None):
    '''
    Returns the normalized betweenness centrality of the nodes and edges.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    btype : str, optional (default 'both')
        The centrality that should be returned (either 'node', 'edge', or
        'both'). By default, both betweenness centralities are computed.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.

    Returns
    -------
    nb : :class:`numpy.ndarray`
        The nodes' betweenness if `btype` is 'node' or 'both'
    eb : :class:`numpy.ndarray`
        The edges' betweenness if `btype` is 'edge' or 'both'

    References
    ----------
    .. [wiki-betw] http://en.wikipedia.org/wiki/Centrality#Betweenness_centrality
    .. [gt-betw] :gtdoc:`centrality.betweenness`
    .. [ig-ebetw] :igdoc:`edge_betweenness`
    .. [ig-nbetw] :igdoc:`betweenness`
    .. [nx-ebetw] :nxdoc:`algorithms.centrality.edge_betweenness_centrality`
    .. [nx-nbetw] :nxdoc:`networkx.algorithms.centrality.betweenness_centrality`
    '''
    raise NotImplementedError(_backend_required)


def subgraph_centrality(graph, weights=True, nodes=None,
                        normalize="max_centrality"):
    '''
    Returns the subgraph centrality for each node in the graph.

    Defined according to [Estrada2005]_ as:

    .. math::

        sc(i) = e^{W}_{ii}

    where :math:`W` is the (potentially weighted and normalized) adjacency
    matrix.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Network to analyze.
    weights : bool or string, optional (default: True)
        Whether weights should be taken into account; if True, then connections
        are weighed by their synaptic strength, if False, then a binary matrix
        is returned, if `weights` is a string, then the ponderation is the
        correponding value of the edge attribute (e.g. "distance" will return
        an adjacency matrix where each connection is multiplied by its length).
    nodes : array-like container with node ids, optional (default = all nodes)
        Nodes for which the subgraph centrality should be returned (all
        centralities are computed anyway in the algorithm).
    normalize : str or False, optional (default: "max_centrality")
        Whether the centrality should be normalized. Accepted normalizations
        are "max_eigenvalue" (the matrix is divided by its largest eigenvalue),
        "max_centrality" (the largest centrality is one), and ``False`` to get
        the non-normalized centralities.

    Returns
    -------
    centralities : :class:`numpy.ndarray`
        The subgraph centrality of each node.

    References
    ----------
    .. [Estrada2005] Ernesto Estrada and Juan A. Rodríguez-Velázquez,
       Subgraph centrality in complex networks, PHYSICAL REVIEW E 71, 056103
       (2005), :doi:`10.1103/PhysRevE.71.056103`, :arxiv:`cond-mat/0504730`.
    '''
    adj_mat = graph.adjacency_matrix(types=False, weights=weights).tocsc()

    centralities = None

    if normalize == "max_centrality":
        centralities = spl.expm(adj_mat / adj_mat.max()).diagonal()
        centralities /= centralities.max()
    elif normalize == "max_eigenvalue":
        norm, _ = spl.eigs(adj_mat, k=1)
        centralities = spl.expm(adj_mat / norm).diagonal()
    elif normalize is False:
        centralities = spl.expm(adj_mat).diagonal()
    else:
        raise InvalidArgument('`normalize` should be either False, "eigenmax",'
                              ' or "centralmax".')

    if nodes is None:
        return centralities

    return centralities[nodes]


# ------------------- #
# Spectral properties #
# ------------------- #

def spectral_radius(graph, typed=True, weights=True):
    '''
    Spectral radius of the graph, defined as the eigenvalue of greatest module.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Network to analyze.
    typed : bool, optional (default: True)
        Whether the excitatory/inhibitory type of the connnections should be
        considered.
    weights : bool, optional (default: True)
        Whether weights should be taken into account, defaults to the "weight"
        edge attribute if present.

    Returns
    -------
    the spectral radius as a float.
    '''
    mat_adj  = graph.adjacency_matrix(types=typed, weights=weights)
    eigenval = []

    try:
        eigenval = spl.eigs(mat_adj, return_eigenvectors=False)
    except spl.eigen.arpack.ArpackNoConvergence as err:
        eigenval = err.eigenvalues

    if len(eigenval):
        return np.amax(np.absolute(eigenval))

    raise spl.eigen.arpack.ArpackNoConvergence()


def adjacency_matrix(graph, types=False, weights=False):
    '''
    Adjacency matrix of the graph.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Network to analyze.
    types : bool, optional (default: False)
        Whether the excitatory/inhibitory type of the connnections should be
        considered (only if the weighing factor is the synaptic strength).
    weights : bool or string, optional (default: False)
        Whether weights should be taken into account; if True, then connections
        are weighed by their synaptic strength, if False, then a binary matrix
        is returned, if `weights` is a string, then the ponderation is the
        correponding value of the edge attribute (e.g. "distance" will return
        an adjacency matrix where each connection is multiplied by its length).

    Returns
    -------
    a :class:`~scipy.sparse.csr_matrix`.

    References
    ----------
    .. [gt-adjacency] :gtdoc:`spectral.adjacency`
    .. [nx-adjacency] :nxdoc:`.convert_matrix.to_scipy_sparse_matrix`
    '''
    return graph.adjacency_matrix(types=types, weights=weights)


# --------------- #
# Node properties #
# --------------- #

def node_attributes(network, attributes, nodes=None, data=None):
    '''
    Return node `attributes` for a set of `nodes`.

    Parameters
    ----------
    network : :class:`~nngt.Graph`
        The graph where the `nodes` belong.
    attributes : str or list
        Attributes which should be returned, among:
        * "betweenness"
        * "clustering"
        * "closeness"
        * "in-degree", "out-degree", "total-degree"
        * "subgraph_centrality"
    nodes : list, optional (default: all nodes)
        Nodes for which the attributes should be returned.
    data : :class:`numpy.array` of shape (N, 2), optional (default: None)
        Potential data on the spike events; if not None, it must contain the
        sender ids on the first column and the spike times on the second.

    Returns
    -------
    values : array-like or dict
        Returns the attributes, either as an array if only one attribute is
        required (`attributes` is a :obj:`str`) or as a :obj:`dict` of arrays.
    '''
    if nonstring_container(attributes):
        values = {}

        for attr in attributes:
            values[attr] = _get_attribute(network, attr, nodes, data)

        return values

    return _get_attribute(network, attributes, nodes, data)


def find_nodes(network, attributes, equal=None, upper_bound=None,
               lower_bound=None, upper_fraction=None, lower_fraction=None,
               data=None):
    '''
    Return the nodes in the graph which fulfill the given conditions.

    Parameters
    ----------
    network : :class:`~nngt.Graph`
        The graph where the `nodes` belong.
    attributes : str or list
        Properties on which the conditions apply, among:
        * "B2" (requires NEST or `data` entry)
        * "betweenness"
        * "clustering"
        * "firing_rate" (requires NEST or `data` entry)
        * "in-degree", "out-degree", "total-degree"
        * "subgraph_centrality"
        * any custom property formerly set by the user
    equal : optional (default: None)
        Value to which `attributes` should be equal. For a given
        property, this entry is cannot be used together with any of the
        others.
    upper_bound : optional (default: None)
        Value which should strictly major `attributes` in the desired
        nodes. Can be combined with all other entries, except `equal`.
    lower_bound : optional (default: None)
        Value which should minor or be equal to the value of `attributes`
        in the desired nodes. Can be combined with all other entries,
        except `equal`.
    upper_fraction : optional (default: None)
        Only the nodes that belong to the `upper_fraction` with the highest
        values for `attributes` are kept.
    lower_fraction : optional (default: None)
        Only the nodes that belong to the `lower_fraction` with the lowest
        values for `attributes` are kept.
    data : :class:`numpy.array` of shape (N, 2), optional (default: None)
        Potential data on the spike events; if not None, it must contain the
        sender ids on the first column and the spike times on the second.

    Notes
    -----
    When combining both `*_fraction` and `*_bound` entries, their effects
    are cumulated, i.e. only the nodes belonging to the fraction AND
    displaying a value that is consistent with the boundary are kept.

    Examples
    --------

        nodes = g.find("in-degree", upper_bound=15, lower_bound=10)
        nodes2 = g.find(["total-degree", "clustering"], equal=[20, None],
            lower=[None, 0.1])
    '''
    if not nonstring_container(attributes):
        attributes = [attributes]
        equal = [equal]
        upper_bound = [upper_bound]
        lower_bound = [lower_bound]
        upper_fraction = [upper_fraction]
        lower_fraction = [lower_fraction]
        assert not np.any([
            len(attributes)-len(equal), len(upper_bound)-len(equal),
            len(lower_bound)-len(equal), len(upper_fraction)-len(equal),
            len(lower_fraction)-len(equal)])

    nodes = set(range(self.node_nb()))

    # find the nodes
    di_attr = node_attributes(self, attributes)
    keep = np.ones(self.node_nb(), dtype=bool)

    for i in range(len(attributes)):
        attr, eq = attributes[i], equal[i]
        ub, lb = upper_bound[i], lower_bound[i]
        uf, lf = upper_fraction[i], lower_fraction[i]
        # check that the combination is valid
        if eq is not None:
            assert (ub is None)*(lb is None)*(uf is None)*(lf is None), \
            "`equal` entry is incompatible with all other entries."
            keep *= (_get_attribute(self, attr) == eq)
        if ub is not None:
            keep *= (_get_attribute(self, attr) < ub)
        if lb is not None:
            keep *= (_get_attribute(self, attr) >= lb)
        values = None
        if uf is not None or lf is not None:
            values = _get_attribute(self, attr)
        if uf is not None:
            num_keep = int(self.node_nb()*uf)
            sort = np.argsort(values)[:-num_keep]
            keep_tmp = np.ones(self.node_nb(), dtype=bool)
            keep_tmp[sort] = 0
            keep *= keep_tmp
        if lf is not None:
            num_keep = int(self.node_nb()*lf)
            sort = np.argsort(values)[:num_keep]
            keep_tmp = np.zeros(self.node_nb(), dtype=bool)
            keep_tmp[sort] = 1
            keep *= keep_tmp

    nodes = nodes.intersection_update(np.array(nodes)[keep])

    return nodes


# ------------- #
# Distributions #
# ------------- #

def degree_distrib(graph, deg_type="total", nodes=None, weights=None,
                   log=False, num_bins='bayes'):
    '''
    Degree distribution of a graph.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        the graph to analyze.
    deg_type : string, optional (default: "total")
        type of degree to consider ("in", "out", or "total").
    nodes : list of ints, optional (default: None)
        Restrict the distribution to a set of nodes (default: all nodes).
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    log : bool, optional (default: False)
        use log-spaced bins.
    num_bins : int, list or str, optional (default: 'bayes')
        Any of the automatic methodes from :func:`numpy.histogram`, or 'bayes'
        will provide automatic bin optimization. Otherwise, an int for the
        number of bins can be provided, or the direct bins list.

    See also
    --------
    :func:`numpy.histogram`, :func:`~nngt.analysis.binning`

    Returns
    -------
    counts : :class:`numpy.array`
        number of nodes in each bin
    deg : :class:`numpy.array`
        bins
    '''
    degrees = graph.get_degrees(deg_type, nodes, weights)

    if num_bins == 'bayes' or is_integer(num_bins):
        num_bins = binning(degrees, bins=num_bins, log=log)

    return np.histogram(degrees, num_bins)


def betweenness_distrib(graph, weights=None, nodes=None, num_nbins='bayes',
                        num_ebins='bayes', log=False):
    '''
    Betweenness distribution of a graph.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        the graph to analyze.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    nodes : list or numpy.array of ints, optional (default: all nodes)
        Restrict the distribution to a set of nodes (only impacts the node
        attribute).
    log : bool, optional (default: False)
        use log-spaced bins.
    num_bins : int, list or str, optional (default: 'bayes')
        Any of the automatic methodes from :func:`numpy.histogram`, or 'bayes'
        will provide automatic bin optimization. Otherwise, an int for the
        number of bins can be provided, or the direct bins list.

    Returns
    -------
    ncounts : :class:`numpy.array`
        number of nodes in each bin
    nbetw : :class:`numpy.array`
        bins for node betweenness
    ecounts : :class:`numpy.array`
        number of edges in each bin
    ebetw : :class:`numpy.array`
        bins for edge betweenness
    '''
    ia_nbetw, ia_ebetw = betweenness(graph, btype="both", weights=weights)

    if nodes is not None:
        ia_nbetw = ia_nbetw[nodes]

    ra_nbins, ra_ebins = None, None

    if num_ebins == 'bayes' or log:
        ra_ebins = binning(ia_ebetw, bins=num_ebins, log=log)
    else:
        ra_ebins = num_ebins

    if num_nbins == 'bayes' or log:
        ra_nbins = binning(ia_nbetw, bins=num_nbins, log=log)
    else:
        ra_nbins = num_nbins

    ncounts, nbetw = np.histogram(ia_nbetw, ra_nbins)
    ecounts, ebetw = np.histogram(ia_ebetw, ra_ebins)

    return ncounts, nbetw, ecounts, ebetw


# ----- #
# Tools #
# ----- #

def binning(x, bins='bayes', log=False):
    """
    Binning function providing automatic binning using Bayesian blocks in
    addition to standard linear and logarithmic uniform bins.

    .. versionadded:: 0.7

    Parameters
    ----------
    x : array-like
        Array of data to be histogrammed
    bins : int, list or 'auto', optional (default: 'bayes')
        If `bins` is 'bayes', in use bayesian blocks for dynamic bin widths; if
        it is an int, the interval will be separated into
    log : bool, optional (default: False)
        Whether the bins should be evenly spaced on a logarithmic scale.
    """
    x = np.asarray(x)
    new_bins = None

    if bins == 'bayes':
        return bayesian_blocks(x)
    elif nonstring_container(bins) or bins == "auto":
        if log:
            ordered = np.sort(x)
            nonzero_min = ordered[ordered > 0][0]
            return np.logspace(np.log10(nonzero_min), np.log10(x.max()), 20)
        return bins
    elif is_integer(bins):
        if log:
            ordered = np.sort(x)
            nonzero_min = ordered[ordered > 0][0]
            return np.logspace(np.log10(nonzero_min), np.log10(x.max()), bins)
        else:
            return np.linspace(x.min(), x.max(), bins)

    raise ValueError("unrecognized bin code: '" + str(bins) + "'.")


def _get_attribute(network, attribute, nodes=None, data=None):
    '''
    If data is not None, must be an np.array of shape (N, 2).
    '''
    if attribute.lower() == "b2":
        return get_b2(network, nodes=nodes, data=data)
    elif attribute == "betweenness":
        betw = network.get_betweenness("node")
        if nodes is not None:
            return betw[nodes]
        return betw
    elif attribute == "closeness":
        return closeness(network, nodes=nodes)
    elif attribute == "clustering":
        return local_clustering(network, nodes=nodes)
    elif "degree" in attribute.lower():
        dtype = attribute[:attribute.index("-")]
        if dtype.startswith("w"):
            return network.get_degrees(
                dtype[1:], nodes=nodes, weights=True)
        else:
            return network.get_degrees(dtype, nodes=nodes)
    elif "strength" in attribute.lower():
        dtype = attribute[:attribute.index("-")]
        return network.get_degrees(dtype, nodes=nodes, weights=True)
    elif attribute == "firing_rate":
        return get_firing_rate(network, nodes=nodes, data=data)
    elif attribute == "subgraph_centrality":
        sc = subgraph_centrality(network)
        if nodes is not None:
            return sc[nodes]
        return sc
    elif attribute in network.node_attributes:
        return network.get_node_attributes(nodes=nodes, name=attribute)

    raise RuntimeError("Attribute '{}' is not available.".format(attribute))


# ------------------------------------ #
# Importing backend-specific functions #
# ------------------------------------ #

if nngt._config["backend"] == "networkx":
    from .nx_functions import *

if nngt._config["backend"] == "igraph":
    from .ig_functions import *

if nngt._config["backend"] == "graph-tool":
    from .gt_functions import *

if nngt._config["backend"] == "nngt":
    from .nngt_functions import *

# update analyze_graph dict
for func in __all__:
    nngt.analyze_graph[func] = locals()[func]



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_generation.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

"""
Test the connect group and type methods.
"""

import pytest
import numpy as np

import nngt
import nngt.generation as ng


@pytest.mark.mpi_skip
def test_fixed():
    ''' Fixed degree with type '''
    pop = nngt.NeuralPop.exc_and_inhib(1000)

    igroup = pop["inhibitory"]
    egroup = pop["excitatory"]

    net = nngt.Network(population=pop)

    deg_e = 50
    ng.connect_neural_types(net, 1, [-1, 1], graph_model="fixed_degree",
                           degree=deg_e, degree_type="out-degree")

    deg_i = 100
    ng.connect_neural_types(net, -1, [-1, 1], graph_model="fixed_degree",
                            degree=deg_i, degree_type="out-degree")

    edeg = net.get_degrees("out", nodes=egroup.ids)
    ideg = net.get_degrees("out", nodes=igroup.ids)

    assert np.all(edeg == deg_e)
    assert np.all(ideg == deg_i)

    edeg = net.get_degrees("in", nodes=egroup.ids)
    ideg = net.get_degrees("in", nodes=igroup.ids)

    avg_deg = (deg_e*egroup.size + deg_i*igroup.size) / pop.size
    std_deg = np.sqrt(avg_deg)

    assert avg_deg - std_deg < np.average(edeg) < avg_deg + std_deg
    assert avg_deg - std_deg < np.average(ideg) < avg_deg + std_deg


@pytest.mark.mpi_skip
def test_gaussian():
    ''' Gaussian degree with groups '''
    pop = nngt.NeuralPop.exc_and_inhib(1000)

    igroup = pop["inhibitory"]
    egroup = pop["excitatory"]

    net = nngt.Network(population=pop)

    avg_e = 50
    std_e = 5
    ng.connect_groups(net, egroup, [igroup, egroup],
                      graph_model="gaussian_degree", avg=avg_e, std=std_e,
                      degree_type="out-degree")

    avg_i = 100
    std_i = 5
    ng.connect_groups(net, igroup, [igroup, egroup],
                      graph_model="gaussian_degree", avg=avg_i, std=std_i,
                      degree_type="out-degree")

    edeg = net.get_degrees("out", nodes=egroup.ids)
    ideg = net.get_degrees("out", nodes=igroup.ids)

    # call only on root process (for mpi) and not if using distributed backend
    if nngt.on_master_process() and nngt.get_config("backend") != "nngt":
        assert avg_e - std_e < np.average(edeg) < avg_e + std_e
        assert avg_i - std_i < np.average(ideg) < avg_i + std_i
    elif nngt.get_config("mpi") and nngt.get_config("backend") == "nngt":
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        num_mpi = comm.Get_size()

        edeg = comm.gather(edeg[rank::num_mpi], root=0)
        ideg = comm.gather(ideg[rank::num_mpi], root=0)

        if nngt.on_master_process():
            assert avg_e - std_e < np.average(edeg) < avg_e + std_e
            assert avg_i - std_i < np.average(ideg) < avg_i + std_i

    edeg = net.get_degrees("in", nodes=egroup.ids)
    ideg = net.get_degrees("in", nodes=igroup.ids)

    avg_deg = (avg_e*egroup.size + avg_i*igroup.size) / pop.size
    std_deg = np.sqrt(avg_deg)

    # call only on root process (for mpi) unless using distributed backend
    if nngt.on_master_process() and nngt.get_config("backend") != "nngt":
        assert avg_deg - std_deg < np.average(edeg) < avg_deg + std_deg
        assert avg_deg - std_deg < np.average(ideg) < avg_deg + std_deg
    elif nngt.get_config("mpi") and nngt.get_config("backend") == "nngt":
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        num_mpi = comm.Get_size()

        edeg = comm.gather(edeg, root=0)
        ideg = comm.gather(ideg, root=0)

        if nngt.on_master_process():
            edeg = np.sum(edeg, axis=0)
            ideg = np.sum(ideg, axis=0)

            assert avg_deg - std_deg < np.average(edeg) < avg_deg + std_deg
            assert avg_deg - std_deg < np.average(ideg) < avg_deg + std_deg


def test_group_vs_type():
    ''' Gaussian degree with groups and types '''
    # first with groups
    nngt.seed(0)

    pop = nngt.NeuralPop.exc_and_inhib(1000)

    igroup = pop["inhibitory"]
    egroup = pop["excitatory"]

    net1 = nngt.Network(population=pop)

    all_groups = list(pop.keys())  # necessary to have same order as types

    avg_e = 50
    std_e = 5
    ng.connect_groups(net1, egroup, all_groups, graph_model="gaussian_degree",
                      avg=avg_e, std=std_e, degree_type="out-degree")

    avg_i = 100
    std_i = 5
    ng.connect_groups(net1, igroup, all_groups, graph_model="gaussian_degree",
                      avg=avg_i, std=std_i, degree_type="out-degree")

    # then with types
    nngt.seed(0)

    pop = nngt.NeuralPop.exc_and_inhib(1000)

    igroup = pop["inhibitory"]
    egroup = pop["excitatory"]

    net2 = nngt.Network(population=pop)

    avg_e = 50
    std_e = 5
    ng.connect_neural_types(net2, 1, [-1, 1], graph_model="gaussian_degree",
                            avg=avg_e, std=std_e, degree_type="out-degree")

    avg_i = 100
    std_i = 5
    ng.connect_neural_types(net2, -1, [-1, 1], graph_model="gaussian_degree",
                            avg=avg_i, std=std_i, degree_type="out-degree")

    # call only on root process (for mpi) unless using distributed backend
    if nngt.on_master_process() or nngt.get_config("backend") == "nngt":
        # check that both networks are equals
        assert np.all(net1.get_degrees() == net2.get_degrees())


if __name__ == "__main__":
    import os

    if os.environ.get("MPI"):
        nngt.set_config("mpi", True)


    if os.environ.get("GL", ""):
        nngt.set_config("backend", os.environ["GL"])

    if not nngt.get_config("mpi"):
        test_fixed()

    test_gaussian()
    test_group_vs_type()



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_analysis.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

""" Test the graph analysis functions """

import os

import numpy as np
import pytest

import nngt
import nngt.analysis as na
import nngt.generation as ng


nngt_backend = (nngt.get_config("backend") == "nngt")

methods = ('barrat', 'continuous', 'onnela', 'zhang')


@pytest.mark.mpi_skip
def test_binary_undirected_clustering():
    '''
    Check that directed local clustering returns undirected value if graph is
    not directed.
    '''
    # pre-defined graph
    num_nodes = 5
    edge_list = [
        (0, 3), (1, 0), (1, 2), (2, 4), (4, 1), (4, 3), (4, 2), (4, 0)
    ]

    # expected results
    loc_clst  = [2/3., 2/3., 1., 1., 0.5]
    glob_clst = 0.6428571428571429

    # create graph
    g = nngt.Graph(nodes=num_nodes)
    g.new_edges(edge_list)

    # check all 3 ways of computing the local clustering
    assert np.all(np.isclose(
        na.local_clustering_binary_undirected(g), loc_clst))

    assert np.all(np.isclose(
        na.local_clustering(g, directed=False), loc_clst))

    assert np.all(np.isclose(
        nngt.analyze_graph["local_clustering"](g, directed=False),
        loc_clst))

    # check all 4 ways of computing the global clustering
    assert np.isclose(
        na.global_clustering(g, directed=False), glob_clst)

    assert np.isclose(na.transitivity(g, directed=False), glob_clst)

    assert np.isclose(
        na.global_clustering_binary_undirected(g), glob_clst)

    assert np.isclose(
        nngt.analyze_graph["global_clustering"](g, directed=False),
        glob_clst)

    # check that self-loops are ignore
    g.new_edge(0, 0, self_loop=True)

    assert np.all(np.isclose(
        na.local_clustering_binary_undirected(g), loc_clst))

    assert np.isclose(
        na.global_clustering_binary_undirected(g), glob_clst)

    # sanity check for local clustering on undirected unweighted graph
    g = ng.erdos_renyi(avg_deg=10, nodes=100, directed=False)

    ccu = na.local_clustering_binary_undirected(g)
    cc  = na.local_clustering(g)

    assert np.all(np.isclose(cc, ccu))


@pytest.mark.mpi_skip
def test_weighted_undirected_clustering():
    '''
    Check relevant properties of weighted clustering:

    * give back the binary definition if all weights are one
    * corner cases for specific networks, see [Saramaki2007]
    * equivalence between no edge and zero-weight edge for 'continuous' method

    Note: onnela and barrat are already check against networkx and igraph
    implementations in libarry_compatibility.py
    '''
    g = ng.erdos_renyi(avg_deg=10, nodes=100, directed=False)

    # recover binary
    ccb = na.local_clustering_binary_undirected(g)

    for method in methods:
        ccw = na.local_clustering(g, weights='weight', method=method)

        assert np.all(np.isclose(ccb, ccw))

    # corner cases
    eps = 1e-15

    # 3 nodes
    num_nodes = 3
    edge_list = [(0, 1), (1, 2), (2, 0)]

    # all epsilon
    weights = [eps, eps, eps]

    g = nngt.Graph(nodes=num_nodes, directed=False)
    g.new_edges(edge_list, attributes={"weight": weights})

    for method in methods:
        cc = na.local_clustering(g, weights='weight', method=method)
        assert np.array_equal(cc, [1, 1, 1])

    # one weight is one
    g.set_weights(np.array([eps, eps, 1]))

    for method in methods:
        cc = na.local_clustering(g, weights='weight', method=method)

        if method == "barrat":
            assert np.all(np.isclose(cc, 1))
        elif method == "zhang":
            assert np.all(np.isclose(cc, [0, 1, 0]))
        else:
            assert np.all(np.isclose(cc, 0, atol=1e-4))

    # two weights are one
    g.set_weights(np.array([eps, 1, 1]))

    for method in methods:
        cc = na.local_clustering(g, weights='weight', method=method)

        if method == "barrat":
            assert np.all(np.isclose(cc, 1))
        elif method == "zhang":
            # due to floating point rounding errors, we get 0.9 instead of 1
            assert np.all(np.isclose(cc, (0.9, 0.9, 0), atol=1e-3))
        else:
            assert np.all(np.isclose(cc, 0, atol=1e-2))

    # 4 nodes
    num_nodes = 4
    edge_list = [(0, 1), (1, 2), (2, 0), (2, 3)]

    g = nngt.Graph(nodes=num_nodes, directed=False)
    g.new_edges(edge_list)

    # out of triangle edge is epsilon
    g.set_weights([1, 1, 1, eps])

    for method in methods:
        cc = na.local_clustering(g, weights='weight', method=method)

        if method == 'barrat':
            assert np.all(np.isclose(cc, [1, 1, 0.5, 0]))
        elif method in ("continuous", "zhang"):
            assert np.all(np.isclose(cc, [1, 1, 1, 0]))
        else:
            assert np.all(np.isclose(cc, [1, 1, 1/3, 0]))

    # out of triangle edge is 1 others are epsilon
    g.set_weights([eps, eps, eps, 1])

    for method in methods:
        cc = na.local_clustering(g, weights='weight', method=method)

        if method == 'barrat':
            assert np.all(np.isclose(cc, [1, 1, 0, 0]))
        else:
            assert np.all(np.isclose(cc, 0))

    # opposite triangle edge is 1 others are epsilon
    g.set_weights([1, eps, eps, eps])

    for method in methods:
        cc = na.local_clustering(g, weights='weight', method=method)

        if method == 'barrat':
            assert np.all(np.isclose(cc, [1, 1, 1/3, 0]))
        elif method == "zhang":
            assert np.all(np.isclose(cc, [0, 0, 1/3, 0]))
        else:
            assert np.all(np.isclose(cc, 0, atol=1e-5))

    # adjacent triangle edge is 1 others are epsilon
    g.set_weights([eps, 1, eps, eps])

    for method in methods:
        cc = na.local_clustering(g, weights='weight', method=method)

        if method == 'barrat':
            assert np.all(np.isclose(cc, [1, 1, 1/2, 0]))
        elif method == "zhang":
            assert np.all(np.isclose(cc, [1, 0, 0, 0]))
        else:
            assert np.all(np.isclose(cc, 0, atol=1e-4))

    # check zero-weight edge/no edge equivalence for continuous method
    num_nodes = 6
    edge_list = [(0, 1), (1, 2), (2, 0), (2, 3), (4, 5)]

    g = nngt.Graph(nodes=num_nodes, directed=False)
    g.new_edges(edge_list)

    g.set_weights([1/64, 1/729, 1/64, 1/729, 1])

    # triangle is 1/20736
    # triplets are [1/64, 1/216, 62/5832, 0, 0, 0]
    expected = [1/324, 1/96, 9/1984, 0, 0, 0]

    cc = na.local_clustering(g, weights='weight', method='continuous')

    assert np.all(np.isclose(cc, expected))

    # 0-weight case
    g.set_weights([1/64, 1/729, 1/64, 0, 1])

    cc0 = na.local_clustering(g, weights='weight', method='continuous')

    # no-edge case
    edge_list = [(0, 1), (1, 2), (2, 0), (4, 5)]

    g = nngt.Graph(nodes=num_nodes, directed=False)
    g.new_edges(edge_list)
    g.set_weights([1/64, 1/729, 1/64, 1])

    expected = [1/324, 1/96, 1/96, 0, 0, 0]

    ccn = na.local_clustering(g, weights='weight', method='continuous')

    assert np.all(np.isclose(cc0, ccn))
    assert np.all(np.isclose(cc0, expected))


@pytest.mark.mpi_skip
def test_weighted_directed_clustering():
    num_nodes = 6
    edge_list = [(0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 3), (4, 5)]

    g = nngt.Graph(nodes=num_nodes)
    g.new_edges(edge_list)

    # continuous
    g.set_weights([1/64, 1/729, 1/64, 1/729, 1/64, 1/729, 1])

    # expected triangles and triplets
    # s_sqrt_tot  = np.array([89/216, 31/108, 51/216, 1/27, 1, 1])
    # s_tot       = np.array([2251/46656, 761/23328, 307/15552, 1/729, 1, 1])
    # s_recip     = np.array([35/1728, 1/64, 1/216, 0, 0, 0])
    # triplets_c  = np.square(s_sqrt_tot) - s_tot - 2*s_recip
    triangles_c = np.array([97/839808, 97/839808, 97/839808, 0, 0, 0])
    triplets_c  = np.array([35/432, 1/54, 13/486, 0, 0, 0])

    assert np.all(np.isclose(
        triangles_c,
        na.triangle_count(g, weights='weight', method='continuous')))

    assert np.all(np.isclose(
        triplets_c,
        na.triplet_count(g, weights='weight', method='continuous')))

    triplets_c[-3:] = 1
    expected = triangles_c / triplets_c

    cc = na.local_clustering(g, weights='weight', method='continuous')

    assert np.all(np.isclose(cc, expected))

    # barrat (clemente version for reciprocal strength)
    g.set_weights([1/4, 1/9, 1/4, 1/9, 1/4, 1/9, 1])

    # d_tot       = np.array([4, 3, 4, 1, 1, 1])
    # s_recip     = np.array([31/72, 1/4, 0, 0, 0])
    # triplets_b  = s_tot*(d_tot - 1) - s_recip
    triangles_b = np.array([31/36, 13/18, 7/12, 0, 0, 0])
    triplets_b  = np.array([31/18, 13/18, 25/18, 0, 0, 0])

    assert np.all(np.isclose(
        triangles_b, na.triangle_count(g, weights='weight', method='barrat')))

    assert np.all(np.isclose(
        triplets_b, na.triplet_count(g, weights='weight', method='barrat')))

    triplets_b[-3:] = 1
    expected = triangles_b / triplets_b

    cc = na.local_clustering(g, weights='weight', method='barrat')

    assert np.all(np.isclose(cc, expected))

    # onnela
    triplets_o  = np.array([8, 4, 10, 0, 0, 0])
    triangles_o = np.array(
        [0.672764902429877, 0.672764902429877, 0.672764902429877, 0, 0, 0])

    assert np.array_equal(
        triplets_o, na.triplet_count(g, weights='weight', method='onnela'))

    assert np.all(np.isclose(
        triangles_o, na.triangle_count(g, weights='weight', method="onnela")))
    
    triplets_o[-3:] = 1
    expected = triangles_o / triplets_o

    cc = na.local_clustering(g, weights='weight', method='onnela')

    assert np.all(np.isclose(cc, expected))

    # zhang
    g.set_weights([1/2, 1/3, 1/2, 1/3, 1/2, 1/3, 1])

    triangles_z = np.array([5/18, 5/18, 5/18, 0, 0, 0])
    triplets_z  = np.array([5/3, 2/3, 4/3, 0, 0, 0])

    assert np.all(np.isclose(
        triplets_z, na.triplet_count(g, weights='weight', method='zhang')))

    assert np.all(np.isclose(
        triangles_z, na.triangle_count(g, weights='weight', method="zhang")))
    
    triplets_z[-3:] = 1
    expected = triangles_z / triplets_z

    cc = na.local_clustering(g, weights='weight', method='zhang')

    assert np.all(np.isclose(cc, expected))


@pytest.mark.mpi_skip
def test_clustering_parameters():
    W = np.array([
        [0, 1, 1, 1.],
        [1, 0, 1, 0],
        [1, 1, 0, 0],
        [1, 0, 0, 0]
    ])

    g = nngt.Graph.from_matrix(W, directed=True)

    cc_bu = na.local_clustering_binary_undirected(g)

    for m in methods:
        # combine
        for combine in ("sum", "max", "min", "mean"):
            cc_wu = na.local_clustering(g, directed=False, weights=True,
                                        combine_weights=combine, method=m)

            assert np.all(np.isclose(cc_wu, cc_bu))

        # subset of nodes
        nodes = [1, [1, 2]]

        for n in nodes:
            cc = na.local_clustering_binary_undirected(g, nodes=n)

            assert np.all(np.isclose(cc, cc_bu[n]))

            cc = na.local_clustering(g, nodes=n, method=m)

            assert np.all(np.isclose(cc, cc_bu[n]))

    # check different but equivalent matrices
    for m in methods:
        W = np.array([
            [0, 1, 1, 0.6],
            [1, 0, 1, 0],
            [1, 1, 0, 0],
            [0.4, 0, 0, 0]
        ])

        g = nngt.Graph.from_matrix(W, directed=True)

        cc_sum = na.local_clustering(g, directed=False, weights="weight",
                                     combine_weights="mean", method=m)

        W = np.array([
            [0, 1, 1, 0.5],
            [1, 0, 1, 0],
            [1, 1, 0, 0],
            [0.5, 0, 0, 0]
        ])

        g = nngt.Graph.from_matrix(W, directed=False)

        cc_sym = na.local_clustering(g, weights="weight", method=m)

        assert np.all(np.isclose(cc_sum, cc_sym))

        # subset of nodes
        nodes = [1, [1, 2]]

        cc_dir = na.local_clustering(g, weights="weight", method=m)

        for n in nodes:
            cc = na.local_clustering(g, nodes=n, weights="weight", method=m)

            assert np.all(np.isclose(cc, cc_dir[n]))


@pytest.mark.mpi_skip
def test_partial_directed_clustering():
    # cycle
    num_nodes = 3
    edge_list = [
        (0, 1), (1, 2), (2, 0)
    ]

    g = nngt.Graph(num_nodes)
    g.new_edges(edge_list)

    for m in methods:
        assert np.array_equal(
            na.local_clustering(g, mode="cycle", method=m), [1, 1, 1])

        assert np.array_equal(
            na.local_clustering(g, mode="cycle", method=m, weights="weight"),
            [1, 1, 1])

    # middleman, fan-in, fan-out
    num_nodes = 3
    edge_list = [
        (0, 1), (0, 2), (1, 2)
    ]

    g = nngt.Graph(num_nodes)
    g.new_edges(edge_list)

    for m in methods:
        assert np.array_equal(
            na.local_clustering(g, mode="cycle", method=m), [0, 0, 0])

        assert np.array_equal(
            na.local_clustering(g, mode="cycle", method=m, weights="weight"),
            [0, 0, 0])

        assert np.array_equal(
            na.local_clustering(g, mode="middleman", method=m), [0, 1, 0])

        assert np.array_equal(
            na.local_clustering(g, mode="middleman", method=m,
                                weights="weight"),
            [0, 1, 0])

        assert np.array_equal(
            na.local_clustering(g, mode="fan-in", method=m), [0, 0, 0.5])

        assert np.array_equal(
            na.local_clustering(g, mode="fan-in", method=m, weights="weight"),
            [0, 0, 0.5])

        assert np.array_equal(
            na.local_clustering(g, mode="fan-out", method=m), [0.5, 0, 0])

        assert np.array_equal(
            na.local_clustering(g, mode="fan-out", method=m, weights="weight"),
            [0.5, 0, 0])

    # complex graph
    g = nngt.Graph(4)
    g.new_edges([(0, 3), (1, 0), (1, 2), (1, 3), (2, 1), (3, 2)])


    for m in methods:
        assert np.array_equal(
            na.local_clustering(g, mode="cycle", method=m), [0, 0.5, 1, 0.5])

        assert np.array_equal(
            na.local_clustering(g, mode="middleman", method=m), [1, 0, 0, 0.5])

        assert np.array_equal(
            na.local_clustering(g, mode="fan-in", method=m), [0, 0, 0.5, 0.5])

        assert np.array_equal(
            na.local_clustering(g, mode="fan-out", method=m), [0, 1/3, 0, 0])

    # weighted (barrat)
    g.set_weights([1, 0.5, 0.4, 1, 0.6, 1])

    assert np.all(np.isclose(
        na.local_clustering(g, mode="fan-out", weights="weight",
                            method="barrat"),
        [0, 1.45/3.8, 0, 0]))

    assert np.array_equal(
        na.local_clustering(g, mode="fan-in", weights="weight",
                            method="barrat"), [0, 0, 0.5, 0.5])

    assert np.all(np.isclose(
        na.local_clustering(g, mode="cycle", weights="weight", method="barrat"),
        [0, 0.8/1.35, 1, 0.5]))

    assert np.array_equal(
        na.local_clustering(g, mode="middleman", weights="weight",
                            method="barrat"), [1, 0, 0, 0.5])

    # weighted (onnela)
    g.set_weights([1, 1/8, 1, 1, 1, 1/8])

    assert np.array_equal(
        na.local_clustering(g, mode="middleman", weights="weight",
                            method="onnela"), [0.5, 0, 0, 0.25])

    assert np.array_equal(
        na.local_clustering(g, mode="cycle", weights="weight", method="onnela"),
        [0, 0.25, 0.5, 0.25])

    assert np.all(np.isclose(
        na.local_clustering(g, mode="fan-out", weights="weight",
                            method="onnela"), [0, 1/6, 0, 0]))

    assert np.array_equal(
        na.local_clustering(g, mode="fan-in", weights="weight",
                            method="onnela"), [0, 0, 0.25, 0.25])

    # weighted (continuous)
    g.set_weights([1, 1/64, 1/64, 1, 1, 1])

    assert np.all(np.isclose(
        na.local_clustering(g, mode="fan-out", weights="weight"),
        [0, 4/17, 0, 0]))

    assert np.all(np.isclose(
        na.local_clustering(g, mode="fan-in", weights="weight"),
        [0, 0, 0.25, 1/32]))

    assert np.all(np.isclose(
        na.local_clustering(g, mode="cycle", weights="weight"),
        [0, 8/9, 1, 0.5]))

    assert np.all(np.isclose(
        na.local_clustering(g, mode="middleman", weights="weight"),
        [0.5, 0, 0, 1/32]))

    # weighted (zhang)
    g.set_weights([1, 0.25, 0.5, 1, 1, 0.5])

    assert np.all(np.isclose(
        na.local_clustering(g, mode="fan-out", weights="weight", method="zhang"),
        [0, 2/7, 0, 0]))

    assert np.all(np.isclose(
        na.local_clustering(g, mode="fan-in", weights="weight", method="zhang"),
        [0, 0, 0.5, 0.125]))

    assert np.all(np.isclose(
        na.local_clustering(g, mode="cycle", weights="weight", method="zhang"),
        [0, 2/5, 1, 0.5]))

    assert np.all(np.isclose(
        na.local_clustering(g, mode="middleman", weights="weight", method="zhang"),
        [1, 0, 0, 0.25]))


@pytest.mark.mpi_skip
def test_global_clustering():
    num_nodes = 3

    g = nngt.Graph(num_nodes)
    g.new_edges([(0, 1), (0, 2), (2, 1)])

    with pytest.raises(AssertionError):
        na.global_clustering(g, method='plop')

    modes   = ('total', 'cycle', 'middleman', 'fan-in', 'fan-out')
    results = (0.5, 0, 1, 0.5, 0.5)

    for m in methods:
        for mode, res in zip(modes, results):
            gc = na.global_clustering(g, method=m, mode=mode)

            assert gc == res


@pytest.mark.mpi_skip
def test_reciprocity():
    ''' Check reciprocity result '''
    num_nodes  = 5
    edge_list1 = [(0, 3), (1, 0), (1, 2), (2, 4), (4, 1), (4, 3), (4, 2)]
    edge_list2 = [(0, 3), (1, 0), (1, 2), (2, 4), (4, 1), (4, 3)]

    recip = 2/7

    # directed
    g = nngt.Graph(nodes=num_nodes, directed=True)
    g.new_edges(edge_list1)

    assert np.isclose(
        nngt.analyze_graph["reciprocity"](g), recip)

    # undirected
    g = nngt.Graph(nodes=num_nodes, directed=False)
    g.new_edges(edge_list2)

    assert nngt.analyze_graph["reciprocity"](g) == 1.


@pytest.mark.mpi_skip
def test_iedges():
    ''' Check the computation of the number of inhibitory edges '''
    num_nodes = 5
    edge_list = [(0, 3), (1, 0), (1, 2), (2, 4), (4, 1), (4, 3)]

    types = [1, -1, 1, 1, -1, 1]

    for directed in (True, False):
        g = nngt.Graph(nodes=num_nodes, directed=directed)
        g.new_edges(edge_list)

        # from list
        g.set_types(types)

        assert na.num_iedges(g) == 2

        # from node list
        nodes = [1, 2]
        num_inhib = 3

        g.set_types(-1, nodes=nodes)

        assert na.num_iedges(g) == 3

        # from edge fraction
        g.set_types(-1, fraction=0.5)

        assert na.num_iedges(g) == 3


@pytest.mark.mpi_skip
@pytest.mark.skipif(nngt_backend, reason="Not implemented")
def test_swp():
    ''' Check small-world propensity '''
    num_nodes = 500
    k_latt = 20

    # SWP for different extreme p values (0, 1)
    expected = 1 - 1/np.sqrt(2)

    weights  = {"distribution": "uniform", "lower": 0.5, "upper": 5}

    for directed in (True, False):
        for w in (None, weights):
            for p in (0, 1):
                use_weights = None if w is None else "weight"

                if use_weights and not p:
                    # for weighted SWP, it only works for the ER graph limit,
                    # not for the lattice limit since the weights are uniform
                    continue

                g = ng.watts_strogatz(k_latt, p, nodes=num_nodes,
                                      directed=directed, weights=w)

                swp = na.small_world_propensity(
                    g, use_diameter=True, weights=use_weights)

                if w is None:
                    assert np.isclose(swp, expected, atol=0.01)
                else:
                    assert np.isclose(swp, expected, atol=0.02)

    # check options for binary only
    g = ng.watts_strogatz(k_latt, 0, nodes=num_nodes, directed=True)

    assert np.isclose(
        na.small_world_propensity(g, use_global_clustering=False),
        expected, atol=0.01)

    assert np.isclose(
        na.small_world_propensity(g, use_diameter=False), expected, atol=0.01)

    # run a random version just to check with remaining edges in the
    # algorithm
    for directed in (False, True):
        g = ng.newman_watts(10, 0.03, nodes=num_nodes, directed=directed)
        na.small_world_propensity(g, use_diameter=True)

    # check that undirected option works
    g = ng.newman_watts(10, 0.03, nodes=num_nodes, weighted=False)
    u = g.to_undirected()

    swp_g = na.small_world_propensity(g, use_diameter=True, directed=False)
    swp_u = na.small_world_propensity(u, use_diameter=True)

    if not (np.isnan(swp_g) or np.isnan(swp_u)):
        # check is necessary as the rewired graph might be unconnected
        assert np.isclose(swp_g, swp_u, 0.02)


@pytest.mark.mpi_skip
def test_local_closure():
    # undirected
    g = nngt.Graph(4, directed=False)

    g.new_edges([(0, 1), (0, 3), (1, 2), (1, 3)])

    assert np.all(np.isclose(na.local_closure(g), [2/3, 1, 0, 2/3]))

    # undirected weighted (normal/zhang)
    g.set_weights([1, 0.5, 0.25, 1])

    assert np.all(np.isclose(
        na.local_closure(g, weights="weight"),
        [4/7, 1, 0, 4/7]))

    # check weight normalization
    g.set_weights(2*np.array([1, 0.5, 0.25, 1]))

    assert np.all(np.isclose(
        na.local_closure(g, weights="weight"),
        [4/7, 1, 0, 4/7]))

    # undirected weighted (continuous)
    g.set_weights([1, 1/64, 1/4, 1])

    assert np.all(np.isclose(
        na.local_closure(g, weights="weight", method="continuous"),
        [1/13, 0.5, 0, 1/13]))

    # directed (circle)
    g = nngt.Graph(3)

    g.new_edges([(0, 1), (1, 2), (2, 0)])

    assert np.array_equal(na.local_closure(g), [1, 1, 1])
    assert np.array_equal(na.local_closure(g, mode="cycle-out"), [1, 1, 1])
    assert np.array_equal(na.local_closure(g, mode="cycle-in"), [1, 1, 1])

    for mode in ("fan-in", "fan-out"):
        assert np.array_equal(na.local_closure(g, mode=mode), [0, 0, 0])
    
    with pytest.raises(ValueError):
        na.local_closure(g, method="plop")
    
    with pytest.raises(ValueError):
        na.local_closure(g, mode="circle")

    # directed (fan-in/out)
    g = nngt.Graph(3)

    g.new_edges([(0, 1), (0, 2), (1, 2)])

    assert np.array_equal(na.local_closure(g, mode="fan-out"), [1, 0, 0])
    assert np.array_equal(na.local_closure(g, mode="fan-in"), [0, 0, 1])

    for mode in ("cycle-in", "cycle-out"):
        assert np.array_equal(na.local_closure(g, mode=mode), [0, 0, 0])

    # directed weighted
    g = nngt.Graph(4)

    g.new_edges([(0, 1), (1, 2), (1, 3), (2, 1), (3, 0), (3, 1)])

    # weights for normal/zhang
    g.set_weights([1, 0.25, 0.5, 0.25, 0.5, 1])

    modes = ("cycle-out", "cycle-in", "fan-in", "fan-out")
    clsrs = [[1/3, 1, 0, 1/3], [1, 0.5, 0, 2/5], [0, 1, 0, 0], [0, 0, 0, 2/3]]

    for mode, res in zip(modes, clsrs):
        assert np.all(np.isclose(
            na.local_closure(g, weights="weight", mode=mode), res))

    # weights for continuous
    g.set_weights([1, 0.25, 1, 0.25, 1/64, 1])
    
    clsrs = [
        [1/24, 0.5, 0, 0.1],
        [0.5, 0.5, 0, 1/24],
        [0, 0.5, 0, 0],
        [0, 0, 0, 0.1]
    ]
    
    for mode, res in zip(modes, clsrs):
        assert np.all(np.isclose(
            na.local_closure(g, weights="weight", method="continuous",
                             mode=mode),
            res))


if __name__ == "__main__":
    if not nngt.get_config("mpi"):
        test_binary_undirected_clustering()
        test_weighted_undirected_clustering()
        test_weighted_directed_clustering()
        test_reciprocity()
        test_iedges()
        test_swp()
        test_partial_directed_clustering()
        test_clustering_parameters()
        test_global_clustering()
        test_local_closure()



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_attributes.py

"""
Test the population and group functions.
"""

import numpy as np
import pytest

import nngt


def test_groups():
    ids = [i for i in range(10)]
    g1 = nngt.NeuralGroup(ids, neuron_type=None)

    assert ids == g1.ids
    assert len(ids) == len(g1)
    assert g1.neuron_model is None
    assert g1.neuron_type is None
    assert not g1.has_model
    assert g1.is_metagroup

    g2 = nngt.NeuralGroup(ids, neuron_type=None)

    assert g1 == g2
    assert g1.name.startswith("MetaGroup ")
    assert int(g1.name[9:]) + 1 == int(g2.name[9:])

    g3 = nngt.NeuralGroup(ids, neuron_type=1, name="test")
    assert g1 != g3
    assert g3.name == "test"
    assert g3.neuron_type == 1

    g4 = nngt.NeuralGroup()

    assert g4.neuron_type == 1


def test_add_nodes():
    g = nngt.Group()

    assert not g.ids

    nodes = list(range(10))

    g.add_nodes(nodes)

    assert nodes == g.ids

    with pytest.raises(ValueError):
        g.add_nodes(10)


def test_population():
    # basic population
    ids1 = [i for i in range(10)]
    g1   = nngt.NeuralGroup(ids1, neuron_type=1)

    ids2 = [i for i in range(10, 20)]
    g2   = nngt.NeuralGroup(ids2, neuron_type=-1)

    pop = nngt.NeuralPop.from_groups((g1, g2), ("g1", "g2"), with_models=False)

    assert "g1" in pop and "g2" in pop
    assert set(ids1 + ids2) == set(pop.ids)
    assert pop.inhibitory == ids2
    assert pop.excitatory == ids1
    assert len(pop) == 2
    assert pop.size == len(ids1) + len(ids2)
    assert pop["g1"].ids == ids1

    #add meta groups
    g3  = nngt.NeuralGroup(ids1[::2] + ids2[::2], neuron_type=None)
    pop.add_meta_group(g3)

    g4 = pop.create_meta_group([i for i in range(pop.size) if i % 3],
                               "mg2")

    assert g3 in pop.meta_groups.values()
    assert g4 in pop.meta_groups.values()

    assert set(g2.ids).issuperset(g3.inhibitory)
    assert set(g1.ids).issuperset(g3.excitatory)

    assert set(g2.ids).issuperset(g4.inhibitory)
    assert set(g1.ids).issuperset(g4.excitatory)

    # build population from empty one
    pop = nngt.NeuralPop()
    g5  = pop.create_group([i for i in range(10)], "group5",
                           neuron_model="none")
    g2.neuron_model = "none"
    pop["group2"] = g2

    pop.set_model({"neuron": "random_model"})

    assert pop.is_valid

    # basic populations
    pop = nngt.NeuralPop.uniform(100)

    assert len(pop) == 1
    assert pop.size == 100

    pop = nngt.NeuralPop.exc_and_inhib(100)

    assert len(pop) == 2
    assert len(pop["excitatory"].ids) == 80
    assert pop["excitatory"].ids == pop.excitatory
    assert len(pop["inhibitory"].ids) == 20
    assert pop["inhibitory"].ids == pop.inhibitory


def test_failed_pop():
    ids1 = [i for i in range(10)]
    g1   = nngt.NeuralGroup(ids1, neuron_type=None)

    ids2 = [i for i in range(10, 20)]
    g2   = nngt.NeuralGroup(ids2, neuron_type=None)

    failed = True

    try:
        pop = nngt.NeuralPop.from_groups((g1, g2), ("g1", "g2"))
        failed = False
    except:
        pass

    try:
        pop = nngt.NeuralPop.from_groups((g1, g2), ("g1", "g2"),
                                         with_models=False)
        failed = False
    except:
        pass

    assert failed


@pytest.mark.mpi_skip
def test_group_structure():
    metagroup = nngt.Group(100, metagroup=True)

    g1 = nngt.Group(50)
    g2 = nngt.Group(80, name="group2")

    struct = nngt.Structure.from_groups((g1, g2))

    with pytest.raises(AssertionError):
        g = nngt.Graph(100, structure=struct)

    struct.create_meta_group(range(20, 70), name="meta2")

    g = nngt.Graph(structure=struct)

    nngt.generation.connect_groups(g, g1, g1, "all_to_all")

    nngt.generation.connect_groups(g, struct, g1, "erdos_renyi", avg_deg=5,
                                   ignore_invalid=True)

    nngt.generation.connect_groups(g, g2, struct, "erdos_renyi", avg_deg=5,
                                   ignore_invalid=True)


if __name__ == "__main__":
    test_groups()
    test_add_nodes()
    test_population()
    test_failed_pop()
    test_group_structure()



#-*- coding:utf-8 -*-
#
# lib/graph_helpers.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy

import numpy as np

from .errors import InvalidArgument
from .rng_tools import _eprop_distribution
from .test_functions import nonstring_container, is_integer


""" Helper functions for graph classes """

def _edge_prop(prop):
    ''' Return edge property `name` as a distribution dict '''
    if is_integer(prop) or isinstance(prop, np.float):
        return {"distribution": "constant", "value": prop}
    elif isinstance(prop, dict):
        return prop.copy()
    elif nonstring_container(prop):
        return {'distribution': 'custom', 'values': prop}
    elif prop is None:
        return {'distribution': 'constant'}
    else:
        raise InvalidArgument("Edge property must be either a dict, a list, or"
                              " a number; received {}".format(type(prop)))


def _get_edge_attr(graph, elist, attribute, prop=None, last_edges=False):
    '''
    Returns the values associated to a given edge attribute.

    Parameters
    ----------
    graph : the graph
    elist : the edges (N, 2)
    attributes : str
        The name of the attribute to set.
    prop : dict or array-like, optional (default: None)
        Properties associated to the `attribute`.

    Returns
    -------
    attr : array-like
        Values of the `attribute`
    '''
    # check the weights
    if "weight" == attribute:
        weights = np.ones(len(elist))

        if graph.is_weighted():
            prop = graph._w if prop is None else _edge_prop(prop)

            params = {
                k: v for (k, v) in prop.items() if k != "distribution"
            }

            weights = _eprop_distribution(
                graph, prop["distribution"], elist=elist,
                last_edges=last_edges, **params)

        # if dealing with network, check inhibitory weight factor
        if graph.is_network() and not np.isclose(graph._iwf, 1.):
            keep = graph.node_attributes['type'][elist[:, 0]] < 0
            weights[keep] *= graph._iwf

        return weights

    # also check delays
    if "delay" == attribute:
        delays = np.ones(len(elist))
        if prop is None and hasattr(graph, "_d"):
            prop = graph._d
        elif prop is not None:
            prop = _edge_prop(prop)
        params = {
            k: v for (k, v) in prop.items() if k != "distribution"
        }
        delays = _eprop_distribution(
            graph, prop["distribution"], elist=elist,
            last_edges=last_edges, **params)

        return delays

    # take care of others
    prop = _edge_prop(prop)
    params = {k: v for (k, v) in prop.items() if k != "distribution"}
    return _eprop_distribution(
        graph, prop["distribution"], elist=elist, last_edges=last_edges,
        **params)


def _get_syn_param(src_name, src_group, tgt_name, tgt_group, syn_spec,
                   key=None):
    '''
    Return the most specific synaptic properties in `syn_spec` with respect to
    connections between `src_group` and `tgt_group`.
    Priority is given to source (presynaptic properties): they come last.
    '''
    group_keys = []

    for k in syn_spec.keys():
        group_keys.extend(k)

    group_keys = set(group_keys)

    src_type = src_group.neuron_type
    tgt_type = tgt_group.neuron_type

    # entry for source type and target type
    dict_prop = syn_spec.get((src_type, tgt_type), {})
    key_prop = dict_prop.get(key, None)

    # entry for source type and target name
    if tgt_name in group_keys:
        dict_prop = syn_spec.get((src_type, tgt_name), dict_prop)
        key_prop = dict_prop.get(key, key_prop)
    # entry for source name and target type
    if src_name in group_keys:
        dict_prop = syn_spec.get((src_name, tgt_type), dict_prop)
        key_prop = dict_prop.get(key, key_prop)
    # entry for source name and target name
    if src_name in group_keys and tgt_name in group_keys:
        dict_prop = syn_spec.get((src_name, tgt_name), dict_prop)
        key_prop = dict_prop.get(key, key_prop)

    if key is not None:
        return deepcopy(key_prop)

    return deepcopy(dict_prop)


def _get_dtype(value):
    if is_integer(value):
        return "int"
    elif isinstance(value, float):
        return "double"
    if isinstance(value, (bytes, str)):
        return "string"

    return "object"


def _post_del_update(graph, nodes, remapping=None):
    '''
    Remove entries in positions and structure after node deletion.

    Parameters
    ----------
    graph : the graph
    nodes : set of nodes that have been deleted
    remapping : dict mapping old to new node ids
    '''
    num_old = graph.node_nb() + len(nodes)

    keep = np.ones(num_old, dtype=bool)
    keep[list(nodes)] = 0

    if graph.is_spatial():
        graph._pos = graph._pos[keep]

    if graph.structure is not None:
        struct = graph.structure

        # update structure
        struct._size = struct._max_id = struct._desired_size = graph.node_nb()

        struct._groups = struct._groups[keep]

        # map old and new node ids
        if remapping is None:
            idx = 0

            remapping = {}

            for n in range(num_old):
                if n not in nodes:
                    remapping[n] = idx
                    idx += 1

        # update groups
        for g in struct.values():
            g._ids = {remapping[n] for n in g._ids if n not in nodes}
            g._desired_size = len(g._ids)

        for m in struct._meta_groups.values():
            m._ids = {remapping[n] for n in m._ids if n not in nodes}
            m._desired_size = len(m._ids)


def _get_matrices(g, directed, weights, weighted, combine_weights,
                  normed=False, exponent=None, remove_self_loops=True):
    '''
    Return the relevant matrices:
    * W, Wu if weighted
    * A, Au otherwise
    '''
    if weighted:
        # weighted undirected
        W  = g.adjacency_matrix(weights=weights)

        if normed:
            W /= W.max()

        # remove potential self-loops
        if remove_self_loops:
            W.setdiag(0.)

        if exponent is not None:
            W = W.power(exponent)

        Wu = W

        if g.is_directed():
            if directed:
                Wu  = W + W.T
            elif not directed:
                if combine_weights == "sum":
                    Wu = W + W.T

                    if normed:
                        Wu /= Wu.max()
                elif combine_weights == "mean":
                    A = g.adjacency_matrix().astype(float)
                    A += A.T

                    Wu = (W + W.T).multiply(A.power(-1))
                elif combine_weights == "max":
                    Wu = W.maximum(W.T)
                elif combine_weights == "min":
                    A = g.adjacency_matrix()

                    s = A.multiply(A.T)
                    a = (A - s).maximum(0)

                    wa = W.multiply(a)

                    Wu = wa + wa.T + W.multiply(s).minimum(W.T.multiply(s))
                else:
                    raise ValueError("Unknown `combine_weights`: '{}'.".format(
                                     combine_weights))

        return W, Wu

    # binary undirected
    A = g.adjacency_matrix()

    # remove potential self-loops
    if remove_self_loops:
        A.setdiag(0.)

    Au = A

    if g.is_directed():
        Au = Au + Au.T

        if not directed:
            Au.data = np.ones(len(Au.data))

    return A, Au


# ------------------ #
# Graph-tool helpers #
# ------------------ #

def _get_gt_weights(g, weights):
    ''' Return the properly formatted weights '''
    if nonstring_container(weights):
        # user-provided array (test must come first since non hashable)
        return g.graph.new_edge_property("double", vals=weights)
    elif weights in g.edge_attributes:
        # existing edge attribute
        return g.graph.edge_properties[weights]
    elif weights is True:
        # "normal" weights
        return g.graph.edge_properties['weight']
    elif not weights:
        # unweighted
        return None

    raise ValueError("Unknown edge attribute '" + str(weights) + "'.")


def _get_gt_graph(g, directed, weights, combine_weights=None,
                  return_all=False):
    ''' Returns the correct graph(view) given the options '''
    import graph_tool as gt
    from graph_tool.stats import label_parallel_edges

    directed = g.is_directed() if directed is None else directed

    weights = "weight" if weights is True else weights
    weights = None if weights is False else weights

    if not directed and g.is_directed():
        if weights is not None:
            u, weights = _to_undirected(g, weights, combine_weights)

            if return_all:
                return u, u.graph, u.edge_attributes[weights]

            return u.graph

        graph = gt.GraphView(g.graph, directed=False)
        graph = gt.GraphView(graph, efilt=label_parallel_edges(graph).fa == 0)

        if return_all:
            return g, graph, None

        return graph

    if return_all:
        return g, g.graph, weights

    return g.graph


# -------------- #
# IGraph helpers #
# -------------- #

def _get_ig_weights(g, weights):
    if nonstring_container(weights):
        # user-provided array (test must come first since non hashable)
        return np.array(weights)
    elif weights in g.edge_attributes:
        # existing edge attribute
        return np.array(g.graph.es[weights])
    elif weights is True:
        # "normal" weights
        return np.array(g.graph.es["weight"])
    elif not weights:
        # unweighted
        return None

    raise ValueError("Unknown edge attribute '" + str(weights) + "'.")


def _get_ig_graph(g, directed, weights, combine_weights=None,
                  return_all=False):
    ''' Returns the correct graph(view) given the options '''
    directed = g.is_directed() if directed is None else directed

    weights = "weight" if weights is True else weights
    weights = None if weights is False else weights

    if not directed and g.is_directed():
        if weights is not None:
            u, weights = _to_undirected(g, weights, combine_weights)

            if return_all:
                return u, u.graph, u.edge_attributes[weights]

            return u.graph

        if return_all:
            return g, g.graph.as_undirected(), None

        return g.graph.as_undirected()

    if return_all:
        return g, g.graph, weights

    return g.graph


# ---------------- #
# Networkx helpers #
# ---------------- #

def _get_nx_weights(g, weights):
    ''' Return the properly formatted weights '''
    if nonstring_container(weights):
        # generate a function that returns the weights
        def f(source, target, attr):
            eid = g.edge_id((source, target))
            return weights[eid]

        return f
    elif weights in g.edge_attributes:
        # existing edge attribute
        return weights
    elif weights is True:
        # "normal" weights
        return 'weight'
    elif not weights:
        # unweighted
        return None

    raise ValueError("Unknown attribute '{}' for `weights`.".format(weights))


def _get_nx_graph(g, directed, weights, combine_weights=None,
                  return_all=False):
    ''' Returns the correct graph(view) given the options '''
    directed = g.is_directed() if directed is None else directed

    weights = "weight" if weights is True else weights
    weights = None if weights is False else weights

    if not directed and g.is_directed():
        if weights is not None:
            u, weights = _to_undirected(g, weights, combine_weights)

            if return_all:
                return u, u.graph, u.edge_attributes[weights]

            return u.graph

        if return_all:
            return g, g.graph.to_undirected(as_view=True), None

        return g.graph.to_undirected(as_view=True)

    if return_all:
        return g, g.graph, weights

    return g.graph


def _to_undirected(g, weights, combine_weights):
    if nonstring_container(weights):
        g = g.copy()
        g.new_edge_attribute("tmp", "double", values=weights)

        weights = "tmp"

    return g.to_undirected(combine_weights), weights



#-*- coding:utf-8 -*-
#
# analysis/clustering.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Tools for directed/weighted clsutering analysis """

import numpy as np

import nngt
from nngt.lib import nonstring_container
from nngt.lib.graph_helpers import _get_matrices


__all__ = [
	"global_clustering",
    "global_clustering_binary_undirected",
    "local_closure",
    "local_clustering",
    "local_clustering_binary_undirected",
    "triplet_count",
    "triangle_count",
]


def global_clustering_binary_undirected(g):
    '''
    Returns the undirected global clustering coefficient.

    This corresponds to the ratio of undirected triangles to the number of
    undirected triads.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    '''
    # Note, this function is overloaded by the library-specific version
    # if igraph, graph-tool, or networkx is used
    triangles = triangle_count(g, weights=None, directed=False)
    triplets  = triplet_count(g, weights=None, directed=False)

    return np.sum(triangles) / np.sum(triplets)


def global_clustering(g, directed=True, weights=None, method="continuous",
                      mode="total", combine_weights="mean"):
    '''
    Returns the global clustering coefficient.

    This corresponds to the ratio of triangles to the number of triplets.
    For directed and weighted cases, see definitions of generalized triangles
    and triplets in the associated functions below.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    directed : bool, optional (default: True)
        Whether to compute the directed clustering if the graph is directed.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    method : str, optional (default: 'continuous')
        Method used to compute the weighted clustering, either 'barrat'
        [Barrat2004]_, 'continuous', 'onnela' [Onnela2005]_, or 'zhang'
        [Zhang2005]_.
    mode : str, optional (default: "total")
        Type of clustering to use for directed graphs, among "total", "fan-in",
        "fan-out", "middleman", and "cycle" [Fagiolo2007]_.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.

    References
    ----------
    .. [gt-global-clustering] :gtdoc:`clustering.global_clustering`
    .. [ig-global-clustering] :igdoc:`transitivity_undirected`
    .. [nx-global-clustering] :nxdoc:`algorithms.cluster.transitivity`
    .. [Barrat2004] Barrat, Barthelemy, Pastor-Satorras, Vespignani. The
        Architecture of Complex Weighted Networks. PNAS 2004, 101 (11).
        :doi:`10.1073/pnas.0400087101`.
    .. [Onnela2005] Onnela, Saramäki, Kertész, Kaski. Intensity and Coherence
        of Motifs in Weighted Complex Networks. Phys. Rev. E 2005, 71 (6),
        065103. :doi:`10.1103/physreve.71.065103`, arxiv:`cond-mat/0408629`.
    .. [Fagiolo2007] Fagiolo. Clustering in Complex Directed Networks.
        Phys. Rev. E 2007, 76 (2), 026107. :doi:`10.1103/PhysRevE.76.026107`,
        :arxiv:`physics/0612169`.
    .. [Zhang2005] Zhang, Horvath. A General Framework for Weighted Gene
        Co-Expression Network Analysis. Statistical Applications in Genetics
        and Molecular Biology 2005, 4 (1). :doi:`10.2202/1544-6115.1128`,
        `PDF <https://dibernardo.tigem.it/files/papers/2008/
        zhangbin-statappsgeneticsmolbio.pdf>`_.

    See also
    --------
    :func:`~nngt.analysis.triplet_count`
    :func:`~nngt.analysis.triangle_count`
    '''
    assert method in ("barrat", "continuous", "onnela", "zhang"), \
        "Unknown method '{}'".format(method)

    # check directivity and weights
    directed *= g.is_directed()
    weighted  = weights not in (False, None)

    if not directed and not weighted:
        return global_clustering_binary_undirected(g)
    elif not weighted:
        # directed clustering
        triangles = triangle_count(g, mode=mode)
        triplets  = triplet_count(g, mode=mode)

        return np.sum(triangles) / np.sum(triplets)

    triangles, triplets = _triangles_and_triplets(g, directed, weights, method,
                                                  mode, combine_weights, None)

    return np.sum(triangles) / np.sum(triplets)


def local_closure(g, directed=True, weights=None, method=None,
                  mode="cycle-out", combine_weights="mean"):
    r'''
    Compute the local closure for each node, as defined in [Yin2019]_ as the
    fraction of 2-walks that are closed.

    For undirected binary or weighted adjacency matrices
    :math:`W = \{ w_{ij} \}`, the normal (or Zhang-like) definition is given
    by:

    .. math::

        H_i^0 = \frac{\sum_{j\neq k} w_{ij} w_{jk} w_{ki}}
                     {\sum_{j\neq k\neq i} w_{ij}w_{jk}}
              = \frac{W^3_{ii}}{\sum_{j \neq i} W^2_{ij}}

    While a continuous version of the local closure is also proposed as:

    .. math::

        H_i = \frac{\sum_{j\neq k} \sqrt[3]{w_{ij} w_{jk} w_{ki}}^2}
                   {\sum_{j\neq k\neq i} \sqrt{w_{ij}w_{jk}}}
            = \frac{\left( W^{\left[ \frac{2}{3} \right]} \right)_{ii}^3}
                   {\sum_{j \neq i} \left( W^{\left[ \frac{1}{2} \right]}
                                    \right)^2_{ij}}

    with :math:`W^{[\alpha]} = \{ w^\alpha_{ij} \}`.

    Directed versions of the local closure where defined as follow for a node
    :math:`i` connected to nodes :math:`j` and :math:`k`:

    * "cycle-out" is given by the pattern [(i, j), (j, k), (k, i)],
    * "cycle-in" is given by the pattern [(k, j), (j, i), (i, k)],
    * "fan-in" is given by the pattern [(k, j), (j, i), (k, i)],
    * "fan-out" is given by the pattern [(i, j), (j, k), (i, k)].

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    directed : bool, optional (default: True)
        Whether to compute the directed clustering if the graph is directed.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    method : str, optional (default: 'continuous')
        Method used to compute the weighted clustering, either 'normal'/'zhang'
        or 'continuous'.
    mode : str, optional (default: "circle-out")
        Type of clustering to use for directed graphs, among "circle-out",
        "circle-in", "fan-in", or "fan-out".
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.

    References
    ----------
    .. [Yin2019] Yin, Benson, and Leskovec. The Local Closure Coefficient: A
        New Perspective On Network Clustering. Proceedings of the Twelfth ACM
        International Conference on Web Search and Data Mining 2019, 303-311.
        :doi:`10.1145/3289600.3290991`, `PDF <https://www.cs.cornell.edu/~arb/
        papers/closure-coefficients-WSDM-2019.pdf>`_.
    '''
    directed *= g.is_directed()
    weighted  = weights not in (False, None)

    mat, numer, denom = None, None, None

    if not directed and g.is_directed():
        _, mat = _get_matrices(g, directed, weights, weighted, combine_weights,
                               normed=True)
    else:
        mat = g.adjacency_matrix(weights=weights).astype(float)
        mat /= mat.max()
        mat.setdiag(0)

    mat2, mat3 = None, None

    if directed:
        # set correct matrix
        if mode.endswith("-in"):
            mat = mat.T

        if method == "continuous" and weights is not None:
            sqmat = mat.sqrt()
            cbmat = mat.power(2/3)

            mat2 = sqmat*sqmat

            if mode in ("cycle-in", "cycle-out"):
                mat3 = cbmat*cbmat*cbmat
            elif mode in ("fan-in", "fan-out"):
                mat3 = cbmat*cbmat*cbmat.T
            else:
                raise ValueError("Unknown `mode`: '" + mode + "'.'")
        elif method in ("normal", "zhang", None):
            mat2 = mat*mat

            if mode in ("cycle-in", "cycle-out"):
                mat3 = mat2*mat
            elif mode in ("fan-in", "fan-out"):
                mat3 = mat2*mat.T
            else:
                raise ValueError("Unknown `mode`: '" + mode + "'.'")
        else:
            raise ValueError("Unknown `method`: '" + method + "'.'")
    else:
        # undirected
        if method == "continuous" and weights is not None:
            sqmat = mat.sqrt()
            cbmat = mat.power(2/3)

            mat2 = sqmat*sqmat
            mat3 = cbmat*cbmat*cbmat
        elif method in ("normal", "zhang", None):
            mat2 = mat*mat
            mat3 = mat2*mat
        else:
            raise ValueError("Unknown `method`: '" + method + "'.'")

    numer = mat3.diagonal()
    denom = mat2.sum(axis=1).A1 - mat2.diagonal()

    denom[denom == 0] = 1

    return numer / denom


def local_clustering_binary_undirected(g, nodes=None):
    r'''
    Returns the undirected local clustering coefficient of some `nodes`.

    .. math::

        C_i = \frac{A^3_{ii}}{d_i(d_i - 1)} = \frac{\Delta_i}{T_i}

    with :math:`A` the adjacency matrix, :math:`d_i` the degree of node
    :math:`i`, :math:`\Delta_i` is the number of triangles, and :math:`T_i` is
    the number of triplets to which :math:`i` belongs.

    If `g` is directed, then it is converted to a simple undirected graph
    (no parallel edges), both directed and reciprocal edges are merged into
    a single edge.

    Parameters
    ----------
    g : :class:`~nngt.Graph`
        Graph to analyze.
    nodes : list, optional (default: all nodes)
        The list of nodes for which the clustering will be returned

    Returns
    -------
    lc : :class:`numpy.ndarray`
        The list of clustering coefficients, on per node.

    References
    ----------
    .. [gt-local-clustering] :gtdoc:`clustering.local_clustering`
    .. [ig-local-clustering] :igdoc:`transitivity_local_undirected`
    .. [nx-local-clustering] :nxdoc:`algorithms.cluster.clustering`
    '''
    # Note, this function is overloaded by the library-specific version
    # if igraph, graph-tool, or networkx is used
    triangles = triangle_count(g, weights=None, nodes=nodes, directed=False)
    triplets  = triplet_count(g, weights=None, nodes=nodes, directed=False)

    if nonstring_container(triangles):
        triplets[triangles == 0] = 1
    elif triangles == 0:
        return 0

    return triangles / triplets


def local_clustering(g, nodes=None, directed=True, weights=None,
                     method="continuous", mode="total", combine_weights="mean"):
    r'''
    Local (weighted directed) clustering coefficient of the nodes, ignoring
    self-loops.

    If no weights are requested and the graph is undirected, returns the
    undirected binary clustering.

    For all weighted cases, the weights are assumed to be positive and they are
    normalized to dimensionless values between 0 and 1 through a division by
    the highest weight.

    The default `method` for weighted networks is based on a modification of
    the proposal in [Zhang2005]_ with:

    .. math::

        C_i = \frac{\sum_{jk} \sqrt[3]{w_{ij} w_{ik} w_{jk}}}
                   {\sum_{j\neq k} \sqrt{w_{ij} w_{ik}}}
            = \frac{\left(W^{\left[\frac{2}{3}\right]}\right)^3_{ii}}
                   {\left(s^{\left[\frac{1}{2}\right]}_i\right)^2 - s_i}

    for undirected networks, with
    :math:`W = \{ w_{ij}\} = \tilde{W} / \max(\tilde{W})` the normalized
    weight matrix, :math:`s_i` the normalized strength of node :math:`i`, and
    :math:`s^{[\frac{1}{2}]}_i = \sum_k \sqrt{w_{ik}}` the strength associated
    to the matrix :math:`W^{[\frac{1}{2}]} = \{\sqrt{w_{ij}}\}`.

    For directed networks, we used the total clustering defined in
    [Fagiolo2007]_ by default, hence the second equation becomes:

    .. math::

        C_i = \frac{\frac{1}{2}\left(W^{\left[\frac{2}{3}\right]}
                    + W^{\left[\frac{2}{3}\right],T}\right)^3_{ii}}
                   {\left(s^{\left[\frac{1}{2}\right]}_i\right)^2
                    - 2s^{\leftrightarrow}_i - s_i}

    with :math:`s^{\leftrightarrow} = \sum_k \sqrt{w_{ik}w_{ki}}` the
    reciprocal strength (associated to reciprocal connections).

    For the other modes, see the generalized definitions in [Fagiolo2007]_.

    Contrary to 'barrat' and 'onnela' [Saramaki2007]_, this method displays
    *all* following properties:

    * fully continuous (no jump in clustering when weights go to zero),
    * equivalent to binary clustering when all weights are 1,
    * equivalence between no-edge and zero-weight edge cases,
    * normalized (always between zero and 1).

    Using either 'continuous' or 'zhang' is recommended for weighted graphs.

    Parameters
    ----------
    g : :class:`~nngt.Graph` object
        Graph to analyze.
    nodes : array-like container with node ids, optional (default = all nodes)
        Nodes for which the local clustering coefficient should be computed.
    directed : bool, optional (default: True)
        Whether to compute the directed clustering if the graph is directed.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    method : str, optional (default: 'continuous')
        Method used to compute the weighted clustering, either 'barrat'
        [Barrat2004]_/[Clemente2018]_, 'continuous', 'onnela' [Onnela2005]_/
        [Fagiolo2007]_, or 'zhang' [Zhang2005]_.
    mode : str, optional (default: "total")
        Type of clustering to use for directed graphs, among "total", "fan-in",
        "fan-out", "middleman", and "cycle" [Fagiolo2007]_.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "sum": equivalent to mean due to weight normalization.

    Returns
    -------
    lc : :class:`numpy.ndarray`
        The list of clustering coefficients, on per node.

    References
    ----------
    .. [Barrat2004] Barrat, Barthelemy, Pastor-Satorras, Vespignani. The
        Architecture of Complex Weighted Networks. PNAS 2004, 101 (11).
        :doi:`10.1073/pnas.0400087101`.
    .. [Clemente2018] Clemente, Grassi. Directed Clustering in Weighted
        Networks: A New Perspective. Chaos, Solitons & Fractals 2018, 107,
        26–38. :doi:`10.1016/j.chaos.2017.12.007`, :arxiv:`1706.07322`.
    .. [Fagiolo2007] Fagiolo. Clustering in Complex Directed Networks.
        Phys. Rev. E 2007, 76, (2), 026107. :doi:`10.1103/PhysRevE.76.026107`,
        :arxiv:`physics/0612169`.
    .. [Onnela2005] Onnela, Saramäki, Kertész, Kaski. Intensity and Coherence
        of Motifs in Weighted Complex Networks. Phys. Rev. E 2005, 71 (6),
        065103. :doi:`10.1103/physreve.71.065103`, :arxiv:`cond-mat/0408629`.
    .. [Saramaki2007] Saramäki, Kivelä, Onnela, Kaski, Kertész. Generalizations
        of the Clustering Coefficient to Weighted Complex Networks.
        Phys. Rev. E 2007, 75 (2), 027105. :doi:`10.1103/PhysRevE.75.027105`,
        :arxiv:`cond-mat/0608670`.
    .. [Zhang2005] Zhang, Horvath. A General Framework for Weighted Gene
        Co-Expression Network Analysis. Statistical Applications in Genetics
        and Molecular Biology 2005, 4 (1). :doi:`10.2202/1544-6115.1128`,
        `PDF <https://dibernardo.tigem.it/files/papers/2008/
        zhangbin-statappsgeneticsmolbio.pdf>`_.

    See also
    --------
    :func:`undirected_binary_clustering`
    :func:`global_clustering`
    '''
    # check directivity and weights
    directed *= g.is_directed()
    weighted  = weights not in (None, False)

    triplets, triangles = None, None

    if not directed and not weighted:
        # undirected binary clustering uses the library method
        return local_clustering_binary_undirected(g, nodes=nodes)
    elif not weighted:
        # directed clustering
        triangles = triangle_count(g, nodes=nodes, mode=mode)
        triplets  = triplet_count(g, nodes, mode=mode).astype(float)
    else:
        triangles, triplets = _triangles_and_triplets(
            g, directed, weights, method, mode, combine_weights, nodes)

    if nonstring_container(triplets):
        triplets[triangles == 0] = 1
    elif triangles == 0:
        return 0

    return triangles / triplets


def triangle_count(g, nodes=None, directed=True, weights=None,
                   method="normal", mode="total", combine_weights="mean"):
    '''
    Returns the number or the strength (also called intensity) of triangles
    for each node.

    Parameters
    ----------
    g : :class:`~nngt.Graph` object
        Graph to analyze.
    nodes : array-like container with node ids, optional (default = all nodes)
        Nodes for which the local clustering coefficient should be computed.
    directed : bool, optional (default: True)
        Whether to compute the directed clustering if the graph is directed.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    method : str, optional (default: 'normal')
        Method used to compute the weighted triangles, either 'normal', where
        the weights are directly used, or the definitions associated to the
        weighted clustering: 'barrat' [Barrat2004]_, 'continuous', 'onnela'
        [Onnela2005]_, or 'zhang' [Zhang2005]_.
    mode : str, optional (default: "total")
        Type of clustering to use for directed graphs, among "total", "fan-in",
        "fan-out", "middleman", and "cycle" [Fagiolo2007]_.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.

    Returns
    -------
    tr : array
        Number or weight of triangles to which each node belongs.

    References
    ----------
    .. [Barrat2004] Barrat, Barthelemy, Pastor-Satorras, Vespignani. The
        Architecture of Complex Weighted Networks. PNAS 2004, 101 (11).
        :doi:`10.1073/pnas.0400087101`.
    .. [Fagiolo2007] Fagiolo. Clustering in Complex Directed Networks.
        Phys. Rev. E 2007, 76, (2), 026107. :doi:`10.1103/PhysRevE.76.026107`,
        :arxiv:`physics/0612169`.
    .. [Onnela2005] Onnela, Saramäki, Kertész, Kaski. Intensity and Coherence
        of Motifs in Weighted Complex Networks. Phys. Rev. E 2005, 71 (6),
        065103. :doi:`10.1103/physreve.71.065103`, :arxiv:`cond-mat/0408629`.
    .. [Zhang2005] Zhang, Horvath. A General Framework for Weighted Gene
        Co-Expression Network Analysis. Statistical Applications in Genetics
        and Molecular Biology 2005, 4 (1). :doi:`10.2202/1544-6115.1128`,
        `PDF <https://dibernardo.tigem.it/files/papers/2008/
        zhangbin-statappsgeneticsmolbio.pdf>`_.
    '''
    directed *= g.is_directed()
    weighted  = weights not in (False, None)

    exponent = None

    if method == "onnela":
        exponent = 1/3
    elif method == "continuous":
        exponent = 2/3

    # get relevant matrices (use directed=False to get both dir/undir mat)
    mat, matsym = _get_matrices(
        g, directed, weights, weighted, combine_weights, exponent=exponent,
        normed=True)

    # if unweighted, adj is mat, adjsym is matsym
    adj, adjsym = mat, matsym

    # for barrat, we need both weighted and binary matrices
    if method == "barrat" and weighted:
        adj, adjsym = _get_matrices(g, directed, None, False, combine_weights)

    return _triangle_count(mat, matsym, adj, adjsym, method, mode, weighted,
                           directed, nodes)


def triplet_count(g, nodes=None, directed=True, weights=None,
                  method="normal", mode="total", combine_weights="mean"):
    r'''
    Returns the number or the strength (also called intensity) of triplets for
    each node.

    For binary networks, the triplets of node :math:`i` are defined as:

    .. math::

        T_i = \sum_{j,k} a_{ij}a_{ik}

    Parameters
    ----------
    g : :class:`~nngt.Graph` object
        Graph to analyze.
    nodes : array-like container with node ids, optional (default = all nodes)
        Nodes for which the local clustering coefficient should be computed.
    directed : bool, optional (default: True)
        Whether to compute the directed clustering if the graph is directed.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    method : str, optional (default: 'continuous')
        Method used to compute the weighted triplets, either 'normal', where
        the edge weights are directly used, or the definitions used for
        weighted clustering coefficients, 'barrat' [Barrat2004]_,
        'continuous', 'onnela' [Onnela2005]_, or 'zhang' [Zhang2005]_.
    mode : str, optional (default: "total")
        Type of clustering to use for directed graphs, among "total", "fan-in",
        "fan-out", "middleman", and "cycle" [Fagiolo2007]_.
    combine_weights : str, optional (default: 'mean')
        How to combine the weights of reciprocal edges if the graph is directed
        but `directed` is set to False. It can be:

        * "sum": the sum of the edge attribute values will be used for the new
          edge.
        * "mean": the mean of the edge attribute values will be used for the
          new edge.
        * "min": the minimum of the edge attribute values will be used for the
          new edge.
        * "max": the maximum of the edge attribute values will be used for the
          new edge.

    Returns
    -------
    tr : array
        Number or weight of triplets to which each node belongs.

    References
    ----------
    .. [Barrat2004] Barrat, Barthelemy, Pastor-Satorras, Vespignani. The
        Architecture of Complex Weighted Networks. PNAS 2004, 101 (11).
        :doi:`10.1073/pnas.0400087101`.
    .. [Fagiolo2007] Fagiolo. Clustering in Complex Directed Networks.
        Phys. Rev. E 2007, 76, (2), 026107. :doi:`10.1103/PhysRevE.76.026107`,
        :arxiv:`physics/0612169`.
    .. [Zhang2005] Zhang, Horvath. A General Framework for Weighted Gene
        Co-Expression Network Analysis. Statistical Applications in Genetics
        and Molecular Biology 2005, 4 (1). :doi:`10.2202/1544-6115.1128`,
        `PDF <https://dibernardo.tigem.it/files/papers/2008/
        zhangbin-statappsgeneticsmolbio.pdf>`_.
    '''
    directed *= g.is_directed()
    weighted  = weights not in (False, None)

    # simple binary cases
    if not weighted or method == "onnela":
        # undirected
        if not directed:
            deg = None

            if g.is_directed():
                _, adjsym = _get_matrices(g, directed, None, False,
                                          combine_weights)

                if nodes is None:
                    deg = adjsym.sum(axis=0).A1
                else:
                    deg = adjsym.sum(axis=0).A1[nodes]
            else:
                deg = g.get_degrees(nodes=nodes)

            if nodes is None or nonstring_container(nodes):
                return (0.5*deg*(deg - 1)).astype(int)

            return 0.5*deg*(deg - 1)

        # directed
        if mode in ("total", "cycle", "middleman"):
            adj = g.adjacency_matrix()

            d_recip = (adj*adj).diagonal()

            if nodes is not None:
                d_recip = d_recip[nodes]

            din  = g.get_degrees("in", nodes=nodes)
            dout = g.get_degrees("out", nodes=nodes)

            if mode == "total":
                dtot = din + dout

                return dtot*(dtot - 1) - 2*d_recip

            return din*dout - d_recip
        else:
            assert mode in ("fan-in", "fan-out"), \
                "Unknown mode '{}'".format(mode)

            deg = g.get_degrees(mode[4:], nodes=nodes)

            return deg*(deg - 1)

    # check method for weighted
    W, Wu, A, Au = None, None, None, None

    if method in ("continuous", "normal", "zhang"):
        # we need only the weighted matrices
        W, Wu = _get_matrices(g, directed, weights, weighted,
                              combine_weights=combine_weights, normed=True)
    elif method == "barrat":
        # we need only the (potentially) directed matrices
        W = g.adjacency_matrix(weights=weights)
        A = g.adjacency_matrix()
    else:
        raise ValueError("`method` must be either 'barrat', 'onnela', "
                         "'zhang', or 'continuous'/'normal' (identical "
                         "options).")

    return _triplet_count_weighted(
        g, W, Wu, A, Au, method, mode, directed, weights, nodes)


# ---------------------------------------------------------- #
# Overwrite binary clusterings with library-specific version #
# ---------------------------------------------------------- #

if nngt._config["backend"] == "networkx":
    from .nx_functions import (global_clustering_binary_undirected,
                               local_clustering_binary_undirected)

if nngt._config["backend"] == "igraph":
    from .ig_functions import (global_clustering_binary_undirected,
                               local_clustering_binary_undirected)

if nngt._config["backend"] == "graph-tool":
    from .gt_functions import (global_clustering_binary_undirected,
                               local_clustering_binary_undirected)


# -------------- #
# Tool functions #
# -------------- #


def _triangles_and_triplets(g, directed, weights, method, mode,
                            combine_weights, nodes):
    ''' Return the triangles and triplets '''
    # weighted clustering
    W, Wu, A, Au = None, None, None, None
    triplets = None

    # check the method to get the relevant matrices
    if method == "continuous":
        W, Wu = _get_matrices(g, directed, weights, True, combine_weights,
                              exponent=2/3, normed=True)

        Wtr, Wtru = _get_matrices(g, directed, weights, True, combine_weights,
                                  normed=True)

        triplets = _triplet_count_weighted(
            g, Wtr, Wtru, A, Au, method, mode, directed, weights, nodes)
    if method == "zhang":
        W, Wu = _get_matrices(g, directed, weights, True, combine_weights,
                              normed=True)

        triplets = _triplet_count_weighted(
            g, W, Wu, A, Au, method, mode, directed, weights, nodes)
    elif method == "onnela":
        W, Wu = _get_matrices(g, directed, weights, True, combine_weights,
                              exponent=1/3, normed=True)

        # onnela uses the binary triplets
        triplets = triplet_count(g, nodes=nodes, directed=directed,
                                 mode=mode, weights=None)
    elif method == "barrat":
        # we need all matrices
        W, Wu = _get_matrices(g, directed, weights, True, combine_weights,
                              normed=True)
        A, Au = _get_matrices(g, directed, None, False, combine_weights)

        triplets = _triplet_count_weighted(
            g, W, Wu, A, Au, method, mode, directed, weights, nodes)

    # get triangles and triplet strength
    triangles = _triangle_count(W, Wu, A, Au, method, mode, weighted=True,
                                directed=directed, nodes=nodes)

    return triangles, triplets


def _triangle_count(mat, matsym, adj, adjsym, method, mode, weighted, directed,
                    nodes):
    '''
    (Un)weighted (un)directed triangle count.
    '''
    tr = None

    if method == "barrat":
        if mode == "total":
            tr = 0.5*(matsym*adjsym*adjsym).diagonal()
        elif mode == "cycle":
            tr = 0.5*(mat*adj*adj + mat.T*adj.T*adj.T).diagonal()
        elif mode == "middleman":
            tr = 0.5*(mat.T*adj*adj.T + mat*adj.T*adj).diagonal()
        elif mode == "fan-in":
            tr = 0.5*(mat.T*adjsym*adj).diagonal()
        elif mode == "fan-out":
            tr = 0.5*(mat*adjsym*adj.T).diagonal()
        else:
            raise ValueError("Unknown mode ''.".format(mode))
    else:
        if not weighted:
            mat, matsym = adj, adjsym
        elif method not in ("continuous", "zhang", "normal", "onnela"):
            raise ValueError("Invalid `method`: '{}'".format(method))

        if mode == "total":
            tr = 0.5*(matsym*matsym*matsym).diagonal()
        elif mode == "cycle":
            tr = (mat*mat*mat).diagonal()
        elif mode == "middleman":
            tr = (mat*mat.T*mat).diagonal()
        elif mode == "fan-in":
            tr = (mat.T*mat*mat).diagonal()
        elif mode == "fan-out":
            tr = (mat*mat*mat.T).diagonal()
        else:
            raise ValueError("Unknown mode ''.".format(mode))

    if nodes is None:
        return tr

    return tr[nodes]


def _triplet_count_weighted(g, mat, matsym, adj, adjsym, method, mode,
                            directed, weights, nodes):
    '''
    triplet count, weighted only.
    '''
    tr = None

    if method == "normal":
        pass
    elif method == "continuous":
        if directed:
            sqmat = mat.sqrt()

            if mode == "total":
                s2_sq_tot = np.square(sqmat.sum(axis=0).A1 +
                                      sqmat.sum(axis=1).A1)
                s_tot     = mat.sum(axis=0).A1 + mat.sum(axis=1).A1
                s_recip   = 2*(sqmat*sqmat).diagonal()

                tr = s2_sq_tot - s_tot - s_recip
            elif mode in ("cycle", "middleman"):
                s_sq_out = sqmat.sum(axis=0).A1
                s_sq_in  = sqmat.sum(axis=1).A1
                s_recip  = (sqmat*sqmat).diagonal()

                tr = s_sq_in*s_sq_out - s_recip
            elif mode in ("fan-in", "fan-out"):
                axis  = 0 if mode == "fan-in" else 1
                s2_sq = np.square(sqmat.sum(axis=axis).A1)
                sgth  = mat.sum(axis=axis).A1

                tr = s2_sq - sgth
            else:
                raise ValueError("Unknown mode ''.".format(mode))
        else:
            sqmat = matsym.sqrt()

            s2_sq = np.square(sqmat.sum(axis=0).A1)
            s     = matsym.sum(axis=0).A1

            tr = 0.5*(s2_sq - s)
    elif method == "zhang":
        if directed:
            mat2 = mat.power(2)

            if mode == "total":
                s2_sq_tot = np.square(mat.sum(axis=0).A1 +
                                      mat.sum(axis=1).A1)
                s_tot     = mat2.sum(axis=0).A1 + mat2.sum(axis=1).A1
                s_recip   = 2*(mat*mat).diagonal()

                tr = s2_sq_tot - s_tot - s_recip
            elif mode in ("cycle", "middleman"):
                s_sq_out = mat.sum(axis=0).A1
                s_sq_in  = mat.sum(axis=1).A1
                s_recip  = (mat*mat).diagonal()

                tr = s_sq_in*s_sq_out - s_recip
            elif mode in ("fan-in", "fan-out"):
                axis  = 0 if mode == "fan-in" else 1
                s2_sq = np.square(mat.sum(axis=axis).A1)
                sgth  = mat2.sum(axis=axis).A1

                tr = s2_sq - sgth
            else:
                raise ValueError("Unknown mode ''.".format(mode))
        else:
            mat2 = matsym.power(2)

            s2_sq = np.square(matsym.sum(axis=0).A1)
            s     = mat2.sum(axis=0).A1

            tr = 0.5*(s2_sq - s)
    elif method == "barrat":
        if directed:
            # specifc definition of the reciprocal strength from Clemente
            if mode == "total":
                s_recip = 0.5*(mat*adj + adj*mat).diagonal()

                dtot = g.get_degrees("total")
                wmax = np.max(g.get_weights())
                stot = g.get_degrees("total", weights=weights) / wmax

                tr = stot*(dtot - 1) - 2*s_recip
            elif mode in ("cycle", "middleman"):
                s_recip = 0.5*(mat*adj + adj*mat).diagonal()
                s_in    = mat.sum(axis=0).A1
                s_out   = mat.sum(axis=1).A1
                d_in    = g.get_degrees("in")
                d_out   = g.get_degrees("out")

                tr = 0.5*(s_in*d_out + s_out*d_in) - s_recip
            elif mode in ("fan-in", "fan-out"):
                axis = 0 if mode == "fan-in" else 1
                sgth = mat.sum(axis=axis).A1
                deg  = g.get_degrees(mode[4:])

                tr = sgth*(deg - 1)
            else:
                raise ValueError("Unknown mode ''.".format(mode))
        elif g.is_directed():
            d = adjsym.sum(axis=0).A1
            s = matsym.sum(axis=0).A1

            tr = 0.5*s*(d - 1)
        else:
            d = g.get_degrees()
            s = matsym.sum(axis=0).A1

            tr = 0.5*s*(d - 1)
    else:
        raise ValueError(
            "Invalid `method` for triplet count: '{}'".format(method))

    if nodes is None:
        return tr

    return tr[nodes]
